#!/bin/sh

# install script for @{IDENT}

# TXBASE	where to install the package
# TXDIR		don't change
# BINDIR	location of the binaries (tx6, txx3, etc)
# DOCDIR	online documents
# MANDIR	where to place the man page
# EXMDIR	where to place the example programs

TXBASE=/usr/local/t3x
TXDIR=$TXBASE/r6
DOCDIR=$TXDIR/doc
BINDIR=/usr/local/bin
MANDIR=/usr/local/man/man7
EXMDIR=/usr/local/share/examples/t3x

TXCM=../tools/txcm

(cd ../classes; for cls in char iostream memory string system ttyctl xmem; do
	if [ ! -f ${cls}.ndx ]; then
		echo $cls: >$cls.ndx
	fi
 done)

install -o bin -g bin -d -m 0755 $TXBASE
install -o bin -g bin -d -m 0755 $TXDIR
install -o bin -g bin -d -m 0755 $TXDIR/bin
install -o bin -g bin -d -m 0755 $TXDIR/classes
install -o bin -g bin -d -m 0755 $TXDIR/lib
install -o bin -g bin -d -m 0755 $DOCDIR
install -o bin -g bin -d -m 0755 $EXMDIR

install -o bin -g bin -C -m 0644 t3x.dir $TXDIR/classes
install -o bin -g bin -C -m 0755 txdpf $TXDIR/bin/txdpf
install -o bin -g bin -C -m 0755 txld $TXDIR/bin/txld
install -o bin -g bin -C -m 0755 txopt $TXDIR/bin/txopt
install -o bin -g bin -C -m 0755 txtrn $TXDIR/bin/txtrn
install -o bin -g bin -C -m 0755 txx $TXDIR/bin/txx
ln -fs $TXDIR/bin/txx $BINDIR/txx
ln -fs $BINDIR/txx $BINDIR/txx3

install -o bin -g bin -C -m 0644 ../compiler/tcode.dir $TXDIR/classes/tcode.dir
touch $TXDIR/classes/tcode.cls
touch $TXDIR/classes/CLASSNDX
(cd ../classes; $TXCM -i char)
(cd ../classes; $TXCM -i iostream)
(cd ../classes; $TXCM -i memory)
(cd ../classes; $TXCM -i string)
(cd ../classes; $TXCM -i system)
(cd ../classes; $TXCM -i ttyctl)
(cd ../classes; $TXCM -i xmem)
install -o bin -g bin -C -m 0644 ../classes/r5c.cls $TXDIR/classes
chown bin:bin $TXDIR/classes/*
chown bin:bin $TXDIR/lib/*

install -o bin -g bin -C -m 0755 ../native/txcg@{TARGET} \
	$TXDIR/bin/txcg@{TARGET}
ln -fs $TXDIR/bin/txcg@{TARGET} $TXDIR/bin/txcg386
install -o bin -g bin -C -m 0755 ../native/txcgexe \
	$TXDIR/bin/txcgexe
ln -fs $TXDIR/bin/txcgexe $TXDIR/bin/txcg86
install -o bin -g bin -C -m 0755 ../native/s86 $TXDIR/bin/s86
ln -fs $TXDIR/bin/s86 $BINDIR/s86
install -o bin -g bin -C -m 0644 ../native/libt3x.a \
	$TXDIR/lib/libt3x.a
install -o bin -g bin -C -m 0644 ../native/t3x.lib $TXDIR/lib/t3x.lib
install -o bin -g bin -C -m 0644 ../native/dos_system.so \
	$TXDIR/lib/dos_system.so
install -o bin -g bin -C -m 0644 ../native/dos_ttyctl.so \
	$TXDIR/lib/dos_ttyctl.so
install -o bin -g bin -C -m 0644 ../native/dos_xmem.so \
	$TXDIR/lib/dos_xmem.so

install -o bin -g bin -C -m 0755 ../tools/tx $TXDIR/bin/tx
ln -fs $TXDIR/bin/tx $BINDIR/tx
ln -fs $BINDIR/tx $BINDIR/tx6
install -o bin -g bin -C -m 0755 ../tools/txcm $TXDIR/bin/txcm
ln -fs $TXDIR/bin/txcm $BINDIR/txcm
install -o bin -g bin -C -m 0755 ../tools/txdump $TXDIR/bin/txdump
ln -fs $TXDIR/bin/txdump $BINDIR/txdump
install -o bin -g bin -C -m 0755 ../tools/txprint $TXDIR/bin/txprint
ln -fs $TXDIR/bin/txprint $BINDIR/txprint
install -o bin -g bin -C -m 0644 ../tools/txprint.rules \
	$TXDIR/txprint.rules
install -o bin -g bin -C -m 0755 ../tools/thelp $TXDIR/bin/thelp
ln -fs $TXDIR/bin/thelp $BINDIR/thelp
install -o bin -g bin -C -m 0755 ../tools/txpc $TXDIR/bin/txpc
ln -fs $TXDIR/bin/txpc $BINDIR/txpc

install -o bin -g bin -C -m 0644 ../docs/README.docs \
	$DOCDIR/README.docs
install -o bin -g bin -C -m 0644 ../docs/README $DOCDIR/README
install -o bin -g bin -C -m 0644 ../docs/Changes $DOCDIR
install -o bin -g bin -C -m 0644 ../docs/BOOT.HOWTO $DOCDIR
install -o bin -g bin -C -m 0644 ../docs/FAQ $DOCDIR
install -o bin -g bin -C -m 0644 ../docs/KNOWNBUGS $DOCDIR
install -o bin -g bin -C -m 0644 ../docs/LICENSE $DOCDIR/LICENSE
install -o bin -g bin -C -m 0644 ../docs/*.html $DOCDIR
install -o bin -g bin -C -m 0644 ../docs/*.sd $DOCDIR
install -o bin -g bin -C -m 0644 ../docs/sdv-config \
	$DOCDIR/sdv-config
install -o bin -g bin -C -m 0644 ../docs/t3x.7 $MANDIR/t3x.7
gzip -f9 $MANDIR/t3x.7

install -o bin -g bin -C -m 0644 ../@{HOST}/txrc \
	$TXDIR/system.txrc
install -o bin -g bin -C -m 0644 ../@{HOST}/txrc.dos \
	$TXDIR/dos.txrc
install -o bin -g bin -C -m 0755 ../@{HOST}/ux $BINDIR/ux3

echo
echo "Don't forget to define T3XDIR = $TXDIR"
echo
echo "You might want to install ../@{HOST}/txrc as ~/.txrc, too"
echo

