/*
 *	LIBSHL/DOS -- dynamic linker routines for T3X under DOS.
 *	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 *
 *	TurboC Version.
 */

#ifdef SHL_SUPPORT

#define MAXSHLSIZE	16384u

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <io.h>
#include <process.h>
#include <alloc.h>
#include <dir.h>
#include <dos.h>

#include "txx.h"


#define NSSYM	128
#define NSHL	16
#define MAXSHLP	16
uint	Shlfd[NSHL];
uint	Shsym[NSSYM][2];
uchar	Symref[NSSYM];

#define MAXARGS	128


void shlinit() {
	int	i;

	for (i=0; i<NSHL; i++) Shlfd[i] = 0;
	for (i=0; i<NSSYM; i++) Shsym[i][0] = 0;
}


t3x_shlopen(p)
char	*p;
{
	int	i, libf;
	char	path[81], *tdir;
	uint	lseg, k;
	char	*libp;

	for (i=0; i<NSHL; i++) {
		libp = MK_FP(Shlfd[i], 0);
		if (Shlfd[i] && !strcmp(&libp[0x81], p)) {
			libp[0x80]++;
			return i;
		}
	}
	for (i=0; i<NSHL; i++) if (Shlfd[i] == 0) break;
	if (i >= NSHL) return -1;
	if ((libf = open(p, 0)) < 0) {
		if (p[0] == '/' || p[0] == '.') return -1;
		if ((tdir = getenv("T3XDIR")) == NULL) return -1;
		strcpy(path, tdir);
		strcat(path, "/lib/");
		strcat(path, p);
		if ((libf = open(path, 0)) < 0) return -1;
	}
	setmode(libf, O_BINARY);
	if (allocmem(MAXSHLSIZE>>4, &lseg) != -1) {
		close(libf);
		return -1;
	}
	libp = MK_FP(lseg, 0);
	libp[0x80] = 1;
	strcpy(&libp[0x81], p);
	k = read(libf, &libp[0x100], MAXSHLSIZE-256-1024);
	close(libf);
	if (k >= MAXSHLSIZE-256-1024) return -1;
	Shlfd[i] = lseg;
	return i;
}


t3x_shlsym(fn, n)
char	*fn;
int	n;
{
	int	i, j, k, nproc;
	uint	off;
	uchar	*libp;

	if (n < 0 || n >= NSHL || Shlfd[n] == NULL) return -1;
	for (i=0; i<NSSYM; i++) if (Shsym[i][0] == 0) break;
	if (i >= NSSYM) return -1;
	libp = MK_FP(Shlfd[n], 0x100+5);
	nproc = libp[0];
	for (j=1; nproc; j += libp[j], nproc--) {
		if (!strcmp(fn, &libp[j+3])) break;
	}
	if (!nproc) return -1;
	off = (libp[j+2]<<8) | libp[j+1];
	for (k=0; k<NSSYM; k++) {
		if (Shsym[k][0] == off && Shsym[k][1] == Shlfd[n])
			return k;
	}
	Shsym[i][0] = off;
	Shsym[i][1] = Shlfd[n];
	Symref[i] = n;
	return i;
}


t3x_shlcall(av, sym)
void	**av;
int	sym;
{
	int	(*f)(void *a2, void *a3, void *a4, void *a5, void *a6,
		 void *a7, void *a8, void *a9, void *a10, void *a11,
		 void *a12, void *a13, void *a14, void *a15, void *a16,
		 void *a17);
	int	seg, r;

	if (sym < 0 || sym >= NSSYM || Shsym[sym][0] == 0) return -1;
	f = MK_FP(Shsym[sym][1], Shsym[sym][0]);
	seg = Shsym[sym][1];
	asm push ds;
	asm mov  ax,seg;
	asm mov  ds,ax;
	r = (*f)(av[2], av[3], av[4], av[5], av[6], av[7], av[8], av[9],
		av[10], av[11], av[12], av[13], av[14], av[15], av[16],
		av[17]);
	asm pop  ds;
	return r;
}


W_shlcall(av16, sym)
ushort	*av16;
int	sym;
{
	void	*av[MAXSHLP+2];
	int	i;
	uint	mask;

	if (av16[0] > MAXSHLP) return -1;
	mask = av16[1];
	for (i=0; i<av16[0]+2; i++) {
		if (i > 1 && mask&1)
			av[i] = MK_FP(FP_SEG(Data), av16[i]);
		else
			av[i] = (void *) ((int)av16[i]); /*JUSTDOIT!*/
		if (i > 1) mask>>=1;
	}
	return t3x_shlcall(av, sym);
}


t3x_shlclose(n) {
	int	i;
	char	*libp;

	if (n < 0 || n >= NSHL || Shlfd[n] == NULL) return -1;
	libp = MK_FP(Shlfd[n], 0x80);
	if (--(*libp) > 0) return 0;
	for (i=0; i < NSSYM; i++)
		if (Symref[i] == n)
			Shsym[i][0] = 0;
	freemem(Shlfd[n]);
	Shlfd[n] = 0;
	return 0;
}


#ifndef LIBRARY
X_shlopen()	{ Acc = t3x_shlopen(ARGP(1,0)); }
X_shlsym()	{ Acc = t3x_shlsym(ARGP(2,1), ARG(2,0)); }
X_shlcall()	{ Acc = W_shlcall(ARGP(2,1), ARG(2,0)); }
X_shlclose()	{ Acc = t3x_shlclose(ARG(1,0)); }
#endif

#endif	/* SHL_SUPPORT */
