/*
 *	LIBSHL -- dynamic linker interface for T3XR6 on FreeBSD.
 *	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#ifdef SHL_SUPPORT

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dlfcn.h>

#include "../compiler/txx.h"


#define NSSYM	256
#define NSHL	32
#define MAXSHLP	16

void	*Shlfd[NSHL];
int	(*Shsym[NSSYM])();
uchar	Symref[NSSYM];


shlinit() {
	int	i;

	for (i=0; i<NSHL; i++) Shlfd[i] = NULL;
	for (i=0; i<NSSYM; i++) Shsym[i] = NULL;
}


t3x_shlopen S1(p)
char	*p;
{
	int	i;
	char	path[256], *tdir;

	for (i=0; i<NSHL; i++) if (Shlfd[i] == NULL) break;
	if (i >= NSHL) return -1;
	if ((Shlfd[i] = dlopen(p, RTLD_LAZY)) == NULL) {
		if (p[0] == '/' || p[0] == '.') return -1;
		if ((tdir = getenv("T3XDIR")) == NULL) return -1;
		strcpy(path, tdir);
		strcat(path, "/lib/");
		strcat(path, p);
		if ((Shlfd[i] = dlopen(path, RTLD_LAZY)) == NULL) return -1;
	}
	return i;
}


t3x_shlsym S2(fn, n)
char	*fn;
int	n;
{
	int	(*f)();
	int	i;

	if (n < 0 || n >= NSHL || Shlfd[n] == NULL) return -1;
	for (i=0; i<NSSYM; i++) if (Shsym[i] == NULL) break;
	if (i >= NSSYM) return -1;
#ifdef __ELF__
	f = dlsym(Shlfd[n], &fn[1]);	/* ELF has no leading _ */
#else
	f = dlsym(Shlfd[n], fn);
#endif
	if (f == NULL) return -1;
	Shsym[i] = f;
	Symref[i] = n;
	return i;
}


t3x_shlcall S2(av, sym)
void	**av;
int	sym;
{
	if (sym < 0 || sym >= NSSYM || Shsym[sym] == NULL) return -1;
	return (*Shsym[sym])(av[2], av[3], av[4], av[5], av[6], av[7], av[8],
		av[9], av[10], av[11], av[12], av[13], av[14], av[15], av[16],
		av[17]);
}


#ifndef LIBRARY

W_shlcall(av16, sym)
ushort	*av16;
int	sym;
{
	void	*av[MAXSHLP+2];
	int	i;
	uint	mask;

	if (av16[0] > MAXSHLP) return -1;
	t3x_cvalist(&av[2], &av16[2], av16[1], av16[0]);
	return t3x_shlcall(av, sym);
}

#endif


t3x_shlclose S1(n) {
	int	i, r;

	if (n < 0 || n >= NSHL || Shlfd[n] == NULL) return -1;
	r = dlclose(Shlfd[n]);
	for (i=0; i < NSSYM; i++)
		if (Symref[i] == n)
			Shsym[i] = NULL;
	Shlfd[n] = NULL;
	return r;
}


#ifndef LIBRARY

X_shlopen()	{ Acc = t3x_shlopen(ARGP(1,0)); }
X_shlsym()	{ Acc = t3x_shlsym(ARGP(2,1), ARG(2,0)); }
X_shlcall()	{ Acc = W_shlcall(ARGP(2,1), ARG(2,0)); }
X_shlclose()	{ Acc = t3x_shlclose(ARG(1,0)); }

#endif

#endif	/* SHL_SUPPORT */
