/*
 *	LIBT3X -- common runtime support routines for T3XR6.
 *	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "../compiler/txx.h"


libinit() {
#ifdef SHL_SUPPORT
	shlinit();
#endif
#ifdef dos
	setmode(0, O_BINARY);
	setmode(1, O_BINARY);
#endif
}


#ifdef LIBRARY

int	Argcount;
char	**Argvec;

int main(argc, argv)
int	argc;
char	**argv;
{
	Argcount = argc;
	Argvec = argv;
	libinit();
	_TXSTART_();
	return 0;
}

#endif


t3x_bpw S0() {
	return BPW;
}


t3x_memcopy S3(n, d, s)
int	n;
char	*d, *s;
{
	memmove(s, d, n);
	return 0;
}


t3x_memcomp S3(n, s2, s1)
int	n;
char	*s2, *s1;
{
	return memcmp(s1, s2, n);
}


t3x_memfill S3(n, c, p)
int	n, c;
char	*p;
{
	int	i;

	for (i=0; i<n; i++) p[i] = c;
	return 0;
}


t3x_memscan S3(l, c, p)
int	l, c;
char	*p;
{
	int	i;

	for (i=0; i<l; i++)
		if (p[i] == c) return i;
	return -1;
}


t3x_pack S2(s, v)
uchar	*s;
UCELL	*v;
{
	int	i;

	for (i=0; v[i]; i++) s[i] = v[i];
	s[i] = 0;
	return (i+BPW)/BPW;
}


t3x_unpack S2(v, s)
UCELL	*v;
uchar	*s;
{
	int	i;

	for (i=0; s[i]; i++) v[i] = s[i];
	v[i] = 0;
	return i;
}


t3x_getarg S3(n, b, i)
int	n;
char	*b;
int	i;
{
	int	k;

	if (i >= Argcount) {
		b[0] = 0;
		return -1;
	}
	k = strlen(Argvec[i]);
	k = k>=n? n-1: k;
	memcpy(b, Argvec[i], k);
	b[k] = 0;
	return k;
}


t3x_getenv S3(n, b, vn)
int	n;
char	*b;
char	*vn;
{
	char	*vv;
	int	k;

	vv = getenv(vn);
	if (vv == NULL) {
		b[0] = 0;
		return -1;
	}
	k = strlen(vv);
	k = k>=n? n-1: k;
	memcpy(b, vv, k);
	b[k] = 0;
	return k;
}


t3x_open S2(mode, file)
short	mode;
char	*file;
{
	int	fd;

	if (mode == 1) {
		fd = creat(file, 0666);
	}
	else if (mode == 0 || mode == 2 || mode == 3) {
		fd = open(file, mode==3? O_WRONLY: mode==2? O_RDWR:
			O_RDONLY);
	}
	else {
		return -1;
	}
	if (fd > 0 && mode == 3) lseek(fd, 0L, 2);
	return fd;
}


#define T_SEEK_SET	0
#define T_SEEK_FWD	1
#define T_SEEK_END	2
#define T_SEEK_BCK	3

t3x_seek S3(o, n, fd)
int	o;
uint	n;
int	fd;
{
	if (o == T_SEEK_SET)
		return lseek(fd, (off_t) n * 1L, SEEK_SET) == -1? -1: 0;
	else if (o == T_SEEK_END)
		return (int) lseek(fd, (off_t) n * -1L, SEEK_END) == -1? -1: 0;
	else if (o == T_SEEK_FWD)
		return (int) lseek(fd, (off_t) n * 1L, SEEK_CUR) == -1? -1: 0;
	else if (o == T_SEEK_BCK)
		return (int) lseek(fd, (off_t) n * -1L, SEEK_CUR) == -1? -1: 0;
}


t3x_close S1(fd) {
	return close(fd);
}


t3x_read S3(n, b, fd)
int	n;
char	*b;
int	fd;
{
	return read(fd, b, n);
}


t3x_write S3(n, b, fd)
int	n;
char	*b;
int	fd;
{
	return write(fd, b, n);
}


uchar *t3x_newline S1(s)
uchar	*s;
{
#ifdef unix
	s[0] = '\n';
	s[1] = 0;
#endif
#ifdef dos
	s[0] = '\r';
	s[1] = '\n';
	s[2] = 0;
#endif
	return s;
}


t3x_remove S1(s)
char	*s;
{
	return unlink(s);
}


t3x_rename S2(sn, so)
char	*sn, *so;
{
	if (close(open(sn, 0)) >= 0) return -1;
	return rename(so, sn);
}


#ifdef unix

#ifdef LIBRARY

t3x_cvalist S4(olist, ilist, bmap, n)
void	**olist;
void	*ilist;
ushort	bmap;
short	n;
{
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(void*));
	else
		memcpy(olist, ilist, n*sizeof(void*));
	return 0;
}

#else	/* !LIBRARY */

t3x_cvalist S4(olist, ilist, bmap, n)
void	**olist;
ushort	*ilist;
ushort	bmap;
short	n;
{
	int	i;

	if (n < 0) {
		n = -n;
		for (i=0; i<n; i++)
			ilist[i] = (ushort) ((uint)olist[i]);
			/*YESIMEANIT!*/
	}
	else {
		for (i=0; i<n; i++) {
			if (bmap&1)
				olist[i] = &Data[ilist[i]];
			else
				olist[i] = (void *) ((uint)ilist[i]);
				/*JUSTDOIT!*/
			bmap>>=1;
		}
	}
	return 0;
}

#endif	/* !LIBRARY */

#endif	/* unix */


#ifdef dos

t3x_cvalist S4(olist, ilist, bmap, n)
int	*olist;
int	*ilist;
ushort	bmap;
short	n;
{
	bmap = bmap;	/*USED*/
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(int));
	else
		memcpy(olist, ilist, n*sizeof(int));
	return 0;
}

#endif	/* dos */


#ifndef LIBRARY

X_sys()		{ Acc = 0; }
X_bpw()		{ Acc = t3x_bpw(); }
X_memcopy()	{ Acc = t3x_memcopy(UARG(3,2), ARGP(3,1), ARGP(3,0)); }
X_memcomp()	{ Acc = t3x_memcomp(UARG(3,2), ARGP(3,1), ARGP(3,0)); }
X_memfill()	{ Acc = t3x_memfill(UARG(3,2), ARG(3,1), ARGP(3,0)); }
X_memscan()	{ Acc = t3x_memscan(UARG(3,2), ARG(3,1), ARGP(3,0)); }
X_pack()	{ Acc = t3x_pack(ARGP(2,1), ARGP(2,0)); }
X_unpack()	{ Acc = t3x_unpack(ARGP(2,1), ARGP(2,0)); }
X_getarg()	{ Acc = t3x_getarg(ARG(3,2), ARGP(3,1), ARG(3,0)); }
X_getenv()	{ Acc = t3x_getenv(ARG(3,2), ARGP(3,1), ARGP(3,0)); }
X_open()	{ Acc = t3x_open(ARG(2,1), ARGP(2,0)); }
X_seek()	{ Acc = t3x_seek(ARG(3,2), UARG(3,1), ARG(3,0)); }
X_close()	{ Acc = t3x_close(ARG(1,0)); }
X_read()	{ Acc = t3x_read(ARG(3,2), ARGP(3,1), ARG(3,0)); }
X_write()	{ Acc = t3x_write(ARG(3,2), ARGP(3,1), ARG(3,0)); }
X_newline()	{ Acc = t3x_newline(ARGP(1,0)) - Data; }
X_remove()	{ Acc = t3x_remove(ARGP(1,0)); }
X_rename()	{ Acc = t3x_rename(ARGP(2,1), ARGP(2,0)); }
X_cvalist()	{ Acc = t3x_cvalist(ARGP(4,3), ARGP(4,2), ARG(4,1),
			ARG(4,0)); }

#endif
