!	T3X runtime class: I/O-streams
!	Copyright (C) 1997,1998,2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.

module iostream();


public class iostream(t3x)

 object	t[t3x];

 var	Filedesc,
	Buffer,
	Flags,
	Length,
	Ptr,
	Lim;

 public const	FREAD	= 0x0001,
		FWRITE	= 0x0002,
		FRDWR	= FREAD|FWRITE,
		FKILLCR	= 0x0004,
		FADDCR	= 0x0008,
		FTRANS	= FKILLCR|FADDCR;
 const		FEOF	= 0x0100,
		FLASTW	= 0x0200;

 public const	SEEK_SET = t3x.SEEK_SET,
		SEEK_FWD = t3x.SEEK_FWD,
		SEEK_END = t3x.SEEK_END,
		SEEK_BCK = t3x.SEEK_BCK;


 public create(fd, bufp, len, mode) do
	Filedesc := fd;
	Buffer := bufp;
	Flags := mode;
	Length := len;
	Ptr := 0;
	Lim := 0;
 end


 public open(name, bufp, len, fl) do
	var	fd, mode, ifl;

	ifl := fl & FRDWR;
	mode := ifl = FREAD-> t3x.OREAD:
		ifl = FWRITE-> t3x.OWRITE:
		ifl = FREAD|FWRITE-> t3x.ORDWR: %1;
	if (mode < 0) return %1;
	fd := t.open(name, mode);
	if (fd < 0) return %1;
	return self.create(fd, bufp, len, fl);
 end


 public flush() do var k;
	if (Flags & FWRITE /\ Flags & FLASTW /\ Ptr) do
		k := t.write(Filedesc, Buffer, Ptr);
		if (k \= Ptr) return %1;
	end
	Ptr := 0;
	Lim := 0;
 end


 public close() do
	if (self.flush() = %1) return %1;
	t.close(Filedesc);
	Flags := 0;
 end


 public wrch(ch) do
	Flags := Flags | FLASTW;
	if (Flags & FADDCR) do
		if (ch = '\n') do
			if (Ptr >= Length /\ self.flush() = %1) return %1;
			Buffer::Ptr := '\r';
			Ptr := Ptr+1;
		end
	end
	if (Ptr >= Length /\ self.flush() = %1) return %1;
	Buffer::Ptr := ch;
	Ptr := Ptr+1;
	return ch;
 end


 public write(bufp, len) do
	var	i, p, l, b;

	Flags := Flags | FLASTW;
	i := 0;
	p := Ptr;
	l := Length;
	b := Buffer;
	while (len) do
		if (p >= l) do
			Ptr := p;
			if (self.flush() = %1) return %1;
			p := Ptr;
		end
		ie (Flags & FADDCR) do
			if (bufp::i = '\n') do
				b::p := '\r'; p := p+1;
				if (p >= l) do
					Ptr := p;
					if (self.flush() = %1) return %1;
					p := Ptr;
				end
			end
			b::p := bufp::i;
			p := p+1;
		end
		else do
			b::p := bufp::i;
			p := p+1;
		end
		len := len-1;
		i := i+1;
	end
	Ptr := p;
	return i;
 end


 public writes(str) do var k;
	k := t.memscan(str, 0, 32767);
	if (k > 32766) return %1;
	return self.write(str, k);
 end


 more() do var k;
	if (Flags & FREAD) do
		k := t.read(Filedesc, Buffer, Length);
		if (k <= 0) do
			Flags := FLags | FEOF;
			if (k < 0) return %1;
		end
		Ptr := 0;
		Lim := k;
		return k;
	end
 end


 public rdch() do var c;
	Flags := Flags & ~FLASTW;
	while (1) do
		if (Flags & FEOF) return %1;
		if (Ptr >= Lim /\ more() < 1) return %1;
		c := Buffer::Ptr;
		Ptr := Ptr+1;
		ie (Flags & FKILLCR)
			if (c \= '\r') leave;
		else
			leave;
	end
	return c;
 end


 _read(bufp, len, ckln) do
	var	i, p, e, b;

	Flags := Flags & ~FLASTW;
	i := 0;
	p := Ptr;
	e := Lim;
	b := Buffer;
	while (len) do
		if (p >= e) do
			Ptr := p;
			if (more() < 1) leave;
			p := Ptr;
			e := Lim;
		end
		ie (Flags & FKILLCR) do
			if (b::p \= '\r') do
				bufp::i := b::p;
				i := i+1;
				len := len-1;
			end
		end
		else do
			bufp::i := b::p;
			i := i+1;
			len := len-1;
		end
		p := p+1;
		if (ckln /\ bufp::(i-1) = '\n') leave;
	end
	if (ckln) bufp::i := 0;
	Ptr := p;
	Lim := e;
	return i;
 end


 public read(bufp, len) return _read(bufp, len, 0);


 public reads(bufp, len) return _read(bufp, len, 1);


 public move(off, how) do var delta;
	ie ((how = SEEK_FWD \/ how = SEEK_BCK) /\ Flags & FLASTW) do
		if (self.flush() = %1) return %1;
	end
	else ie (how = SEEK_FWD \/ how = SEEK_BCK) do
		delta := Lim - Ptr;
		if (self.flush() = %1) return %1;
		off := off - delta;
	end
	else do
		if (self.flush() = %1) return %1;
	end
	return t.seek(Filedesc, off, how);
 end


 public eof() return (Flags & FEOF) \= 0;

 public reset() Flags := Flags & ~FEOF;

end


do
end

