!	T3X runtime extension: SYSTEM interface
!	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.

module system();


public class system(t3x)

 object	t[t3x];

 var	SO;
 var	sys_init,
 	sys_pipe,
 	sys_dup,
 	sys_dup2,
 	sys_rdchk,
 	sys_chdir,
 	sys_mkdir,
 	sys_rmdir,
 	sys_getdir,
 	sys_opendir,
 	sys_readdir,
 	sys_closedir,
 	sys_stat,
 	sys_spawn,
 	sys_fork,
 	sys_wait,
 	sys_kill,
 	sys_time,
	sys_fini;

 public const	SPAWN_WAIT = 1,
		SPAWN_NOWAIT = 0;

 public const	SIGTEST = 0,
		SIGTERM = 15,
		SIGKILL = 9;

 public struct STATBUF =
		ST_DEV,
		ST_INO,
		ST_MODE,
		ST_NLINK,
		ST_UID,
		ST_GID,
		ST_RDEV,
		ST_SIZE,
		ST_EXT64,
		ST_MTIME,
		ST_MT_2,
		ST_MT_3,
		ST_MT_4;

 public const	FM_RDOK = 0x01,
		FM_WROK = 0x02,
		FM_EXOK = 0x04,
		FM_ISDIR= 0x08;


 length(s) return t.memscan(s,0,32767);

 abort(msg, name) do
	t.write(2,packed"SYSTEM: ",8);
	t.write(2,msg,length(msg));
	t.write(2,packed": ",2);
	t.write(2,name,length(name));
	t.write(2,packed"\r\n",2);
	halt -1;
 end


 call0(p) return t.shlcall(p, [0,0]);

 call1(p, a) return t.shlcall(p, [1, 0, (a)]);

 call1v(p, a) return t.shlcall(p, [1, 1, (a)]);

 call2(p, a1, a2) return t.shlcall(p, [2, 0, (a1), (a2)]);


 link(name, psym) do var sym;
	sym := t.shlsym(SO, name);
	if (sym < 0) abort(packed"cannot find procedure", name);
	psym[0] := sym;
 end


 public init() do var r, tty_init;
	SO := t.shlopen(packed"system.so");
	if (SO < 0)
		abort(packed"cannot open SYSTEM module", packed"system.so");
	link(packed"_sys_init", @sys_init);
	call0(sys_init);
	link(packed"_sys_pipe", @sys_pipe);
	link(packed"_sys_dup", @sys_dup);
	link(packed"_sys_dup2", @sys_dup2);
	link(packed"_sys_rdchk", @sys_rdchk);
	link(packed"_sys_chdir", @sys_chdir);
	link(packed"_sys_mkdir", @sys_mkdir);
	link(packed"_sys_rmdir", @sys_rmdir);
	link(packed"_sys_getdir", @sys_getdir);
	link(packed"_sys_opendir", @sys_opendir);
	link(packed"_sys_readdir", @sys_readdir);
	link(packed"_sys_closedir", @sys_closedir);
	link(packed"_sys_stat", @sys_stat);
	link(packed"_sys_spawn", @sys_spawn);
	link(packed"_sys_fork", @sys_fork);
	link(packed"_sys_wait", @sys_wait);
	link(packed"_sys_kill", @sys_kill);
	link(packed"_sys_time", @sys_time);
 end


 public pipe(fds) do var r, a[3*2];
	t.cvalist(2, 0, fds, a);
	r := call1v(sys_pipe, a);
	t.cvalist(-2, 0, fds, a);
	return r;
 end


 public dup(fd) do
	return call1(sys_dup, fd);
 end


 public dup2(old, new) do
	return call2(sys_dup2, old, new);
 end


 public rdchk(fd) do
	return call1(sys_rdchk, fd);
 end


 public chdir(s) do
	return call1v(sys_chdir, s);
 end


 public mkdir(s) do
	return call1v(sys_mkdir, s);
 end


 public rmdir(s) do
	return call1v(sys_rmdir, s);
 end


 public getdir(s, n) do
	return t.shlcall(sys_getdir, [2, 1, (s), (n)]);
 end


 public opendir(s) do
	return call1v(sys_opendir, s);
 end


 public readdir(dd, buf, lim) do
	return t.shlcall(sys_readdir, [3, 2, (dd), (buf), (lim)]);
 end


 public closedir(dd) do
	return call1(sys_closedir, dd);
 end


 public stat(path, sb) do var r, i, a[STATBUF];
	r := t.shlcall(sys_stat, [2, 3, (path), (a)]);
	sb[ST_DEV] := a::1 << 8 | a::0;
	sb[ST_INO] := a::3 << 8 | a::2;
	sb[ST_MODE] := a::5 << 8 | a::4;
	sb[ST_NLINK] := a::7 << 8 | a::6;
	sb[ST_UID] := a::9 << 8 | a::8;
	sb[ST_GID] := a::11 << 8 | a::10;
	sb[ST_RDEV] := a::13 << 8 | a::12;
	sb[ST_SIZE] := a::15 << 8 | a::14;
	sb[ST_EXT64] := a::17 << 8 | a::16;
	t.memcopy(@sb[ST_MTIME], @a::18, 8);
	return r;
 end


 public spawn(prog, args, mode) do
	var	abuf[17*2], n, m;

	n := 0;
	m := 0;
	while (args[n]) do
		ie (m)
			m := m<<1 | 1;
		else
			m := 1;
		n := n+1;
	end
	t.cvalist(n+1, m, args, abuf);
	return t.shlcall(sys_spawn, [3, 3, (prog), (abuf), (mode)]);
 end


 public fork() do
	return call0(sys_fork);
 end


 public wait() do
	return call0(sys_wait);
 end


 public kill(pid, sig) do
	return call2(sys_kill, pid, sig);
 end


 public time(tb) do
	return call1v(sys_time, tb);
 end


 public fini() do
	return t.shlclose(SO);
 end

end


do
end
