!	TXDPF -- Tcode3 Dead Procedure Filter
!	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.
!
!	SYNTAX: TXDPF <INFILE >OUTFILE

module txdpf(t3x, tcode);

object	t[t3x];

const	BUFLEN		= 1024;
const	MAXLABEL	= 4096;
const	MAXREF		= 8191;

const	UNUSED	= -1, LEAF = -2;

var	Inbuf::BUFLEN, Ilen, Iptr;
var	Outbuf::BUFLEN, Optr;
var	Infile, Outfile;
var	Procs[MAXLABEL], Uflag::MAXLABEL;
var	CRefs;
var	Entrypoint;
var	Skip;
var	Tmpname::256;

struct	REF = Rdest, Rnext;

var	Refs[REF*MAXREF], Rtop;


length(s) do var k;
	k := 0;
	while (s::k) k := k+1;
	return k;
end


init() do var i;
	ie (t.getarg(1, Tmpname, 256) > 1) do
		ie (Tmpname::0 = '-' /\ Tmpname::1 = 'T') do
			ie (Tmpname::2)
				t.memcopy(Tmpname, @Tmpname::2,
					length(@Tmpname::1));
			else if (t.getarg(2, Tmpname, 256) < 2)
				Tmpname::0 := 0;
		end
		else do
			Tmpname::0 := 0;
		end
	end
	else do
		Tmpname::0 := 0;
	end
	if (\Tmpname::0 /\ t.getenv(packed"TMP", Tmpname, 256) < 0) do
		Tmpname::0 := '.';
		Tmpname::1 := 0;
	end
	for (i=0, 256) if (\Tmpname::i) leave;
	t.memcopy(@Tmpname::i, packed"/__txdpf.tmp", 13);
	Ilen := 0;
	Iptr := 0;
	Optr := 0;
	Infile := 0;
	Outfile := 1;
	for (i=0, MAXLABEL) Procs[i] := UNUSED;
	Crefs := UNUSED;
	Rtop := 0;
	Entrypoint := UNUSED;
	Skip := 0;
end


writep(f, s) t.write(f, s, length(s));


nl(f) do var b::3;
	writep(f, t.newline(b));
end


var	ntoa_buf::32;
ntoa(v, w) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
		w := w-1;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
		w := w-1;
        end
	while (w>0) do
		ntoa_buf::i := '\s'; i := i-1;
		w := w-1;
	end
        return @ntoa_buf::(i+1);
end


error(m) do
	writep(T3X.SYSERR, packed"TXDPF: ");
	writep(T3X.SYSERR, m);
	nl(T3X.SYSERR);
	t.close(t.open(packed"T3X.ERR", 0));
	halt 1;
end


rdch() do
	if (Iptr >= Ilen) do
		if (Ilen < 0) error(packed"unexpected EOF");
		Ilen := t.read(Infile, Inbuf, BUFLEN);
		if (Ilen < 1) do
			Ilen := -1;
			return -1;
		end
		Iptr := 0;
	end
	Iptr := Iptr+1;
	return Inbuf::(Iptr-1);
end


flush() do
	if (t.write(Outfile, Outbuf, Optr) \= Optr)
		error(packed"file write error");
	Optr := 0;
end


wrch(c) do
	if (Skip) return 0;
	if (Optr >= BUFLEN) flush();
	Outbuf::Optr := c;
	Optr := Optr+1;
end


next(op) do var a, i;
	wrch(op);
	if (op & 128) do
		a := rdch() | (rdch()<<8);
		wrch(a); wrch(a>>8);
		if (	op = tcode.IINCL \/ op = tcode.IINCG \/
			op = tcode.IINCI \/
			op = tcode.IINIT \/
			op = tcode.IPUB \/ op = tcode.IEXT \/
			op = tcode.ILSYM \/ op = tcode.IGSYM \/
			op = tcode.IISYM
		) do
			a := rdch() | (rdch()<<8);
			wrch(a); wrch(a>>8);
		end
		if (	op = tcode.ISTR \/ op = tcode.IPSTR \/
			op = tcode.IPUB \/ op = tcode.IEXT \/
			op = tcode.IGSYM \/ op = tcode.ILSYM \/
			op = tcode.IISYM
		) do
			for (i=0, a) wrch(rdch());
		end
	end
	return rdch();
end


newref(p, a) do var n;
	if (Rtop+REF >= REF*MAXREF) error(packed"out of space");
	n := @Refs[Rtop];
	Rtop := Rtop+REF;
	n[Rdest] := a;
	n[Rnext] := p[0];
	p[0] := n;
	return n;
end


addref(p, a) do var q;
	q := p[0];
	while (q \= UNUSED /\ q \= LEAF) do
		if (q[Rdest] = a) return 0;
		q := q[Rnext];
	end
	newref(p, a);
end


phase1() do var op, a, plab, prev;
	Infile := 0;
	Outfile := t.open(Tmpname, T3X.OWRITE);
	if (Outfile < 0) error(packed"could not create tmp file");
	plab := -1;
	op := rdch();
	prev := -1;
	while (op >= 0) do
		if (prev \= -1 /\ op = tcode.IHDR) do
			plab := prev;
			if (plab >= MAXLABEL)
				error(packed"procedure label too big");
			Procs[plab] := LEAF;
		end
		prev := -1;
		ie (	op = tcode.ICALL \/ op = tcode.ILDLAB \/
			op = tcode.ICREF \/ op = tcode.ICLAB
		) do
			wrch(op);
			a := rdch() | (rdch()<<8);
			if (a >= MAXLABEL)
				error(packed"label too big");
			wrch(a); wrch(a>>8);
			ie (op = tcode.ICLAB) do
				ie (a = Entrypoint)
					plab := a;
				else
					prev := a;
			end
			else ie (op = tcode.ICREF /\ plab = -1) do
				addref(@CRefs, a);
			end
			else do
				if (plab >= MAXLABEL)
					error(packed"procedure label too big");
				if (plab = -1)
					error(packed"illegal reference");
				addref(@Procs[plab], a);
			end
			op := rdch();
		end
		else ie (op = tcode.IINIT) do
			wrch(op);
			a := rdch() | (rdch()<<8);
			wrch(a); wrch(a>>8);
			if (a \= 3) error(packed"unsupported Tcode version");
			a := rdch() | (rdch()<<8);
			wrch(a); wrch(a>>8);
			if (Entrypoint \= UNUSED)
				error(packed"multiple entrypoints");
			Entrypoint := a;
			op := rdch();
		end
		else do
			if (op = tcode.IEND) plab := -1;
			op := next(op);
		end
	end
	flush();
	t.close(Outfile);
end


mark(p) do var q, r;
	if (Uflag::p) return 0;
	Uflag::p := 1;
	q := Procs[p];
	while (q \= UNUSED /\ q \= LEAF) do
		mark(q[Rdest]);
		q := q[Rnext];
	end
end


phase2() do var i, p;
	for (i=0, MAXLABEL) Uflag::i := 0;
	if (Entrypoint \= UNUSED) mark(Entrypoint);
	p := CRefs;
	while (p \= UNUSED /\ p \= LEAF) do
		mark(p[Rdest]);
		p := p[Rnext];
	end
end


phase3() do var op, a;
	Ilen := 0;
	Iptr := 0;
	Infile := t.open(Tmpname, T3X.OREAD);
	if (Infile < 0) error(packed"could not reopen tmp file");
	Outfile := 1;
	op := rdch();
	while (op >= 0) do
		ie (op = tcode.ICLAB) do
			a := rdch() | (rdch()<<8);
			op := rdch();
			ie (op = tcode.IHDR)
				Skip := \Uflag::a;
			else if (op = tcode.IEND)
				Skip := 0;
			wrch(tcode.ICLAB);
			wrch(a); wrch(a>>8);
			op := next(op);
		end
		else do
			ie (op = tcode.IEND) do
				op := next(op);
				Skip := 0;
			end
			else do
				op := next(op);
			end
		end
	end
	t.close(Infile);
	flush();
	t.remove(Tmpname);
end


dumprefs() do var i, p;
	for (i=0, MAXLABEL) do
		if (Procs[i] \= UNUSED) do
			writep(T3X.SYSOUT, Uflag::i-> packed"*": packed"\s");
			writep(T3X.SYSOUT, ntoa(i,5));
			writep(T3X.SYSOUT, packed": ");
			p := Procs[i];
			while (p \= UNUSED /\ p \= LEAF) do
				writep(T3X.SYSOUT, ntoa(p[Rdest],0));
				if (p[Rnext] \= UNUSED /\ p[Rnext] \= LEAF)
					writep(T3X.SYSOUT, packed",");
				p := p[Rnext];
			end
			nl(T3X.SYSOUT);
		end
	end
	p := CRefs;
	writep(T3X.SYSOUT, packed" CREFs: ");
	while (p \= -1) do
		writep(T3X.SYSOUT, ntoa(p[Rdest],0));
		if (p[Rnext] \= UNUSED /\ p[Rnext] \= LEAF)
			writep(T3X.SYSOUT, packed",");
		p := p[Rnext];
	end
	nl(T3X.SYSOUT);
	writep(T3X.SYSOUT, packed"  INIT: ");
	writep(T3X.SYSOUT, ntoa(Entrypoint,0));
	nl(T3X.SYSOUT);
end


do
	init();
	phase1();
	phase2();
	phase3();
	!dumprefs();
end
