!	T3X Code Generator Generator
!	Copyright (C) 1998,1999 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.

module txcgg(t3x);

object	t[t3x];

!	---------- I/O-Stream routines ----------

struct IOS =	IOS_FD,
		IOS_BUFFER,
		IOS_FLAGS,
		IOS_LEN,
		IOS_PTR,
		IOS_END;

const		IOF_READ	= 00001,
		IOF_WRITE	= 00002,
		IOF_EOF		= 00004,
		IOF_LASTW	= 00008;

const		SEEK_SET	= 0,
		SEEK_REL	= 1,
		SEEK_END	= 2;


ios_create(iostream, fd, buffer, len, mode) do
	iostream[IOS_FD] := fd;
	iostream[IOS_BUFFER] := buffer;
	iostream[IOS_FLAGS] := mode;
	iostream[IOS_LEN] := len;
	iostream[IOS_PTR] := 0;
	iostream[IOS_END] := 0;
	return iostream;
end


ios_open(iostream, name, buffer, len, flags) do
	var	fd, mode;

	mode := flags = IOF_READ-> T3X.OREAD:
		flags = IOF_WRITE-> T3X.OWRITE:
		flags = IOF_READ|IOF_WRITE-> T3X.ORDWR: %1;
	if (mode < 0) return %1;
	fd := t.open(name, mode);
	if (fd < 0) return %1;
	return ios_create(iostream, fd, buffer, len, flags);
end


ios_flush(iostream) do
	var	k;

	if (	iostream[IOS_FLAGS] & IOF_WRITE /\
		iostream[IOS_FLAGS] & IOF_LASTW /\
		iostream[IOS_PTR]
	) do
		k := t.write(iostream[IOS_FD], iostream[IOS_BUFFER],
			iostream[IOS_PTR]);
		if (k \= iostream[IOS_PTR]) return %1;
	end
	iostream[IOS_PTR] := 0;
	iostream[IOS_END] := 0;
	return 0;
end


ios_close(iostream) do
	if (ios_flush(iostream) = %1) return %1;
	t.close(iostream[IOS_FD]);
	iostream[IOS_FLAGS] := 0;
	return 0;
end


ios_wrch(iostream, ch) do
	iostream[IOS_FLAGS] := iostream[IOS_FLAGS] | IOF_LASTW;
	if (	iostream[IOS_PTR] >= iostream[IOS_LEN] /\
		ios_flush(iostream) = %1
	)
		return %1;
	iostream[IOS_BUFFER]::iostream[IOS_PTR] := ch;
	iostream[IOS_PTR] := iostream[IOS_PTR]+1;
	return ch;
end


ios_write(iostream, buffer, len) do
	var	i, p, l, b;

	iostream[IOS_FLAGS] := iostream[IOS_FLAGS] | IOF_LASTW;
	i := 0;
	p := iostream[IOS_PTR];
	l := iostream[IOS_LEN];
	b := iostream[IOS_BUFFER];
	while (len) do
		if (p >= l) do
			iostream[IOS_PTR] := p;
			if (ios_flush(iostream) = %1) return %1;
			p := iostream[IOS_PTR];
			l := iostream[IOS_LEN];
		end
		b::p := buffer::i;
		p := p+1;
		i := i+1;
		len := len-1;
	end
	iostream[IOS_PTR] := p;
	return i;
end


ios_writep(iostream, str) do var k;
	k := 0;
	while (str::k) k := k+1;
	return ios_write(iostream, str, k);
end


ios__more(iostream) do
	var	k;

	if (iostream[IOS_FLAGS] & IOF_READ) do
		k := t.read(iostream[IOS_FD], iostream[IOS_BUFFER],
			iostream[IOS_LEN]);
		if (k <= 0) do
			iostream[IOS_FLAGS] := iostream[IOS_FLAGS] | IOF_EOF;
			if (k < 0) return %1;
		end
		iostream[IOS_END] := k;
		iostream[IOS_PTR] := 0;
	end
	return k;
end


ios_rdch(iostream) do
	var	c;

	iostream[IOS_FLAGS] := iostream[IOS_FLAGS] & ~IOF_LASTW;
	if (iostream[IOS_FLAGS] & IOF_EOF) return %1;
	if (	iostream[IOS_PTR] >= iostream[IOS_END] /\
		ios__more(iostream) < 1
	)
		return %1;
	c := iostream[IOS_BUFFER]::iostream[IOS_PTR];
	iostream[IOS_PTR] := iostream[IOS_PTR]+1;
	return c;
end


ios__read(iostream, buffer, len, ckln) do
	var	i, p, e, b;

	iostream[IOS_FLAGS] := iostream[IOS_FLAGS] & ~IOF_LASTW;
	i := 0;
	p := iostream[IOS_PTR];
	e := iostream[IOS_END];
	b := iostream[IOS_BUFFER];
	while (len) do
		if (p >= e) do
			iostream[IOS_PTR] := p;
			if (ios__more(iostream) < 1) leave;
			p := iostream[IOS_PTR];
			e := iostream[IOS_END];
		end
		buffer::i := b::p;
		p := p+1;
		i := i+1;
		len := len-1;
		if (ckln /\ buffer::(i-1) = '\n') leave;
	end
	if (ckln) buffer::i := 0;
	iostream[IOS_PTR] := p;
	iostream[IOS_END] := e;
	return i;
end


ios_reads(iostream, buffer, len) return ios__read(iostream, buffer, len, 1);


ios_eof(iostream) return (iostream[IOS_FLAGS] & IOF_EOF) -> %1: 0;

!	---------- End of I/O-Stream routines ----------


const	BUFLEN		= 1024;
const	POOLSIZE	= 16384;

var	In[IOS], Out[IOS], Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Lineno;
var	Pool::POOLSIZE, Pt;


length(s) do var k;
	k := 0;
	while (s::k) k := k+1;
	return k;
end


ucase(s) do
	var	i;

	i := 0;
	while (s::i) do
		if ('a' <= s::i /\ s::i <= 'z')
			s::i := s::i+'A'-'a';
		i := i+1;
	end
end


writep(f, s) t.write(f, s, length(s));


nl(f) do var b::3;
	writep(f, t.newline(b));
end


var	ntoa_buf::32;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
        end
        return @ntoa_buf::(i+1);
end


error(m, n) do
	writep(T3X.SYSERR, packed"TXCGG: ");
	writep(T3X.SYSERR, ntoa(Lineno));
	writep(T3X.SYSERR, packed": ");
	writep(T3X.SYSERR, m);
	if (n) do
		writep(T3X.SYSERR, packed": ");
		writep(T3X.SYSERR, n);
	end
	nl(T3X.SYSERR);
	halt 1;
end


init() do
	if (ios_create(In, 0, Inbuf, BUFLEN, IOF_READ) = %1)
		error(packed"cannot establish input stream", 0);
	if (ios_create(Out, 1, Outbuf, BUFLEN, IOF_WRITE) = %1)
		error(packed"cannot establish output stream", 0);
	Lineno := 0;
	Pt := 0;
end


isname(x) return 'A' <= x /\ x <= 'Z' \/ x = '_';


skip(s, i) do
	while (s::i = '\s' \/ s::i = '\t') i := i+1;
	return i;
end


newmacro(name, text) do
	var	len, k, lt, mac;

	k := length(name);
	lt := text-> length(text): 0;
	len := k+1 + lt+2 + 2;
	if (Pt + len >= POOLSIZE) error(packed"out of macro space", 0);
	mac := Pt;
	Pool::(Pt+2) := k;
	t.memcopy(@Pool::(Pt+3), name, k);
	k := k+3;
	Pool::(Pt+k) := lt;
	Pool::(Pt+k+1) := lt>>8;
	k := k+2;
	if (text) t.memcopy(@Pool::(Pt+k), text, lt);
	Pool::mac := len;
	Pool::(mac+1) := len>>8;
	Pt := Pt+len;
	return mac;
end


addtomacro(mac, text, len) do
	var	k, l;

	if (Pt + len >= POOLSIZE) error(packed"out of string space", 0);
	t.memcopy(@Pool::Pt, text, len);
	Pt := Pt+len;
	k := Pool::(mac+1)<<8 | Pool::mac;
	k := k + len;
	Pool::mac := k;
	Pool::(mac+1) := k>>8;
	l := mac + Pool::(mac+2) + 3;
	k := Pool::(l+1)<<8 | Pool::l;
	k := k + len;
	Pool::l := k;
	Pool::(l+1) := k>>8;
end


truncatemac(mac, len) do
	var	p, k, l;

	k := Pool::(mac+1)<<8 | Pool::mac;
	p := mac + Pool::(mac+2)+3;
	l := Pool::(p+1)<<8 | Pool::p;
	if (l < len) return 0;
	if (Pool::(p+2+l-len) = '\n' /\ Pool::(p+1+l-len) = '\r')
		len := len+1;
	l := l - len;
	Pool::p := l;
	Pool::(p+1) := l>>8;
	k := k - len;
	Pool::mac := k;
	Pool::(mac+1) := k>>8;
	Pt := Pt-len;
end


readdesc() do
	var	line::129;
	var	k, i, j, mac, off;

	k := ios_reads(In, line, 128);
	while (\ios_eof(In)) do
		Lineno := Lineno+1;
		if (k>1 /\ line::(k-2) = '\r') k:=k-1;
		line::(k-1) := 0;
		if (line::0 = 0 \/ line::0 = '!') do
			k := ios_reads(In, line, 128);
			loop;
		end
		if (\isname(line::0)) error(packed"label expected", line);
		i := 0;
		while (isname(line::i)) i := i+1;
		j := i;
		i := skip(line, i);
		if (\line::i) error(packed"missing text", line);
		line::j := 0;
		ie (line::i \= '{') do
			newmacro(line, @line::i);
		end
		else do
			mac := newmacro(line, 0);
			k := ios_reads(In, line, 128);
			off := 0;
			off := skip(line, off);
			while (line::0 \= '}') do
				if (ios_eof(In))
					error(packed"EOF in routine text", 0);
				addtomacro(mac, @line::off, k-off);
				k := ios_reads(In, line, 128);
				off := 0;
			end
			truncatemac(mac, 1);
		end
		k := ios_reads(In, line, 128);
	end
end


dumpmacs() do
	var	p, q, k;

	p := 0;
	while (p<Pt) do
		k := Pool::(p+2);
		t.write(T3X.SYSOUT, @Pool::(p+3), k);
		writep(T3X.SYSOUT, packed" = ");
		q := p+k+3;
		k := Pool::(q+1)<<8 | Pool::q;
		t.write(T3X.SYSOUT, @Pool::(q+2), k);
		nl(T3X.SYSOUT);
		p := p + (Pool::(p+1)<<8 | Pool::p);
	end
end


findmac(s) do
	var	p, q, k, l;

	p := 0;
	l := length(s);
	while (p<Pt) do
		k := Pool::(p+2);
		q := p+k+3;
		if (k = l /\ \t.memcomp(@Pool::(p+3), s, k))
			return @Pool::q;
		p := p + (Pool::(p+1)<<8 | Pool::p);
	end
	return 0;
end


require_bool(s) do
	var	text, k;

	text := findmac(s);
	if (\text) error(packed"undefined symbol", s);
	k := text::1<<8 | text::0;
	ucase(@text::2);
	text := @text::2;
	if (	(k \= 4 \/ t.memcomp(text, packed"TRUE", 4)) /\
		(k \= 5 \/ t.memcomp(text, packed"FALSE", 5))
	)
		error(packed"must be either TRUE or FALSE", s);
	ios_writep(Out, packed"const\s");
	ios_write(Out, s, length(s)); ios_writep(Out, packed"\s=\s");
	ios_wrch(Out, k=4-> '1': '0'); ios_writep(Out, packed";\n");
end


generate() do
	var	text, k;
	var	ch;
	var	name::256, i;

	ios_close(In);
	if (ios_open(In, packed"txcg_frame", Inbuf, BUFLEN, IOF_READ) = %1)
		error(packed"cannot open TXCG frame file", packed"txcg_frame");
	require_bool(packed"SRC_DEST");
	require_bool(packed"LONG_BRANCHES");
	ios_wrch(Out, '\n');
	ch := ios_rdch(In);
	while (\ios_eof(In)) do
		ie (ch = '{') do
			ch := ios_rdch(In);
			i := 0;
			while (ch \= '}') do
				if (i >= 254) do
					name::255 := 0;
					error(packed"routine name too long",
						name);
				end
				name::i := ch;
				i := i+1;
				ch := ios_rdch(In);
			end
			name::i := 0;
			text := findmac(name);
			if (\text) error(packed"undefined routine", name);
			k := text::1<<8 | text::0;
			ios_write(Out, @text::2, k);
		end
		else do
			ios_wrch(Out, ch);
		end
		ch := ios_rdch(In);
	end
end


finish() do
	ios_close(In);
	ios_close(Out);
end


do
	init();
	readdesc();
	!dumpmacs();
	generate();
	finish();
end
