module thelp(t3x, iostream, ttyctl, string, system);

object	t[t3x], str[string], tty[ttyctl], sys[system];


const   SEGLEN  =       24576;
const	BUFLEN	=	1024;
const	SBUFLEN	=	256;
const	NLINES	=	100;
const	NCOLS	=	80;
const	MAXCOLS	=	NCOLS;
const   MAXNDX  =       64;
const   MAXNDX2 =       256;
const	MAXNAMES=	4096;
const	MAXFILES=	NLINES+%4;

const	NEXTTOPIC = 256, PREVTOPIC = 257, LIST_KEYWORDS = 258;
struct	MODES = M_TOP, M_INDEX, M_VIEW, M_KEY;

struct LOCATION =
	L_hseg,
	L_lseg,
	L_offset,
	L_name;


object	Stream[iostream];
var	Streambuf::BUFLEN;
var	Filelist[MAXFILES], Nfiles;
var	File::SEGLEN;
var     Names::MAXNAMES, Nt, Nbase;
var	Buffer::NCOLS*NLINES, Size;
var	Index[MAXNDX2*LOCATION], Nkw;
var	Topindex[MAXNDX*LOCATION], Ntkw;
var	Title::NCOLS, Section, Topic;
var	Lines, Cols;
var	Lineset;
var	Lowseg, Highseg;
var	Message;
var	C_text, C_high;
var	Keyword::128, Keyndx[NLINES], Keyndx2[NLINES];
var	Startdir::SBUFLEN;
var	Directory::128;


init() do var k;
	sys.init();
	tty.init();
	tty.clear();
	tty.mode(1);
	Message := 0;
	Section := 0;
	Topic := 0;
	Lineset := packed"|---++++#.";
	Keyword::0 := 0;
	k := t.getenv(packed"T3XDIR", Directory, 123);
	if (k > 0) str.copy(@Directory::k, packed"/doc");
	Message := 0;
	C_text := 0x17;
	C_high := 0x71;
	Lines := 24;
	Cols := 80;
	Nfiles := 0;
	Nt := 0;
	Nbase := 0;
	sys.getdir(Startdir, SBUFLEN);
end


fini() do
	tty.color(ttyctl.F_GREY|ttyctl.B_BLACK);
	tty.clear();
	tty.fini();
	sys.chdir(Startdir);
	sys.fini();
end


fatal(m, n) do
	fini();
	t.write(T3X.SYSERR, packed"\r\nTHELP ERROR: ", 15);
	t.write(T3X.SYSERR, m, str.length(m));
	if (n) do
		t.write(T3X.SYSERR, packed": " , 2);
		t.write(T3X.SYSERR, n, str.length(n));
	end
	t.write(T3X.SYSERR, packed"\r\n\n", 3);
	halt 1;
end




msg(m, action) do
	if (action = 1) Message := m;
	tty.move(0, Lines-1);
	tty.color(C_high);
	tty.clreol();
	tty.move((Cols/2-1)-(str.length(m))/2, Lines-1);
	tty.writes(m);
end


savename(s) do var k, q;
	k := str.length(s);
	if (Nt+k+1 >= MAXNAMES) fatal(packed"Name pool overflow", s);
	q := @Names::Nt;
	str.copy(q, s);
	Nt := Nt+k+1;
	return q;
end


trim(s, k) do var len;
	len := str.length(s);
	if (len > k) do
		s := s+len-k;
		t.memcopy(s, packed"...", 3);
	end
	return s;
end


gauge(s, len, n, lim) do var i, j;
	n := (n=lim)-> 10: n * 10 / lim;
	for (j=str.length(s), len) s::j := '\s';
	for (i=0, 10) s::(i+j) := Lineset::(i<n->8: 9);
	s::(i+j) := 0;
end


showtitle(pos) do
	var	buf::256, s;
	var	k, q;

	if (Message) return msg(Message, 1);
	str.copy(buf, Title);
	k := str.length(Title);
	if (Section) do
		str.copy(@buf::k, packed" : ");
		k := k + 3;
		q := Section;
		while (	q::0 = '\s' \/ q::0 = ')' \/
			'0' <= q::0 /\ q::0 <= '9'
		)
			q := @q::1;
		str.copy(@buf::k, q);
		k := k + str.length(q);
		if (Topic) do
			str.copy(@buf::k, packed" : ");
			k := k + 3;
			q := Topic;
			while (	q::0 = '\s' \/ q::0 = '.' \/
				'0' <= q::0 /\ q::0 <= '9'
			)
				q := @q::1;
			str.copy(@buf::k, q);
			k := k + str.length(q);
		end
	end
	tty.move(0, Lines-1);
	tty.color(C_high);
	tty.clreol();
	k := Cols-1-12;
	s := trim(buf, k);
	if (pos >= 0) gauge(s, k+2, pos, Size);
	tty.writes(s);
end


position(hi, lo) do
	Stream.move(0, iostream.SEEK_SET);
	while (hi) do
		Stream.move(16384, iostream.SEEK_FWD);
		Stream.move(16384, iostream.SEEK_FWD);
		Stream.move(16384, iostream.SEEK_FWD);
		Stream.move(16384, iostream.SEEK_FWD);
		hi := hi-1;
	end
	Stream.move(lo, iostream.SEEK_FWD);
end


setlocation(high, low) do
	const	LINELEN = 82;
	var	line::LINELEN;
	var	k, len;

	if (Highseg = high /\ Lowseg = low) return 0;
	msg(packed"Loading...", 0);
	Stream.reset();
	position(high, low);
	Highseg := high;
	Lowseg := low;
	len := 0;
	Stream.reads(line, LINELEN);
	while (\Stream.eof()) do
		k := str.length(line);
		if (len + k >= SEGLEN-2) leave;
		line::(k-1) := 0;
		if (line::(k-2) = '\r') do
			line::(k-2) := 0;	! DOS CR
			k := k-1;
		end
		t.memcopy(@File::len, line, k);
		len := len+k;
		Stream.reads(line, LINELEN);
	end
	File::len := 0;
	File::(len+1) := 0;
	showtitle(-1);
end


loadfile(name) do
	const	LINELEN = 82;
	var	line::LINELEN, p, q;
	var	totop, k, top, lastndx;
	var	nsep;
	var	o;

	if (Stream.open(name, StreamBuf, BUFLEN, iostream.FREAD) = %1)
		fatal(packed"Cannot open file", name);
	msg(packed"Building index...", 0);
	Nkw := 0;
	Ntkw := 0;
	Nt := Nbase;
	top := 0;
	nsep := 0;
	Title::0 := 0;
	Lowseg := 0;
	Highseg := 0;
	lastndx := 0;
	totop := 0;
	Stream.reads(line, LINELEN);
	while (\Stream.eof()) do
		k := str.length(line);
		line::(k-1) := 0;
		if (top + k >= SEGLEN-2) do
			o := Lowseg;
			Lowseg := Lowseg + lastndx;
			if (o .>= Lowseg) Highseg := Highseg+1;
			top := top - lastndx;
			o := totop-> @Topindex[(Ntkw-1)*LOCATION]:
				@Index[(Nkw-1)*LOCATION];
			o[L_hseg] := Highseg;
			o[L_lseg] := Lowseg;
			o[L_offset] := o[L_offset] - lastndx;
		end
		t.memcopy(@File::top, line, k+1);
		ie (k < 2) do
			nsep := nsep+1;
		end
		else do
			ie (\Title::0) do
				p := @File::top;
				while (p::0 = '\t') p := @p::1;
				str.copy(Title, p);
			end
			else if (nsep > 1) do
				p := @File::top;
				while (p::0 = '\t') p := p+1;
				ie (p::1 = ')' \/ p::2 = ')') do
					q := Ntkw * LOCATION;
					if (Ntkw >= MAXNDX) fatal(
						packed"Top index overflow", 0);
					Topindex[q+L_hseg] := Highseg;
					Topindex[q+L_lseg] := Lowseg;
					Topindex[q+L_offset] := top;
					Topindex[q+L_name] := savename(p);
					Ntkw := Ntkw+1;
					lastndx := top;
					totop := 1;
				end
				else do
					q := Nkw * LOCATION;
					if (Nkw >= MAXNDX2) fatal(
						packed"Index overflow", 0);
					Index[q+L_hseg] := Highseg;
					Index[q+L_lseg] := Lowseg;
					Index[q+L_offset] := top;
					Index[q+L_name] := savename(p);
					Nkw := Nkw+1;
					lastndx := top;
					totop := 0;
				end
			end
			nsep := 0;
		end
		top := top+k;
		Stream.reads(line, LINELEN);
	end
	File::top := 0;
	File::(top+1) := 0;
	showtitle(-1);
	setlocation(0, 0);
end


clear() t.memfill(Buffer, '\s', NLINES*NCOLS);


build_subindex(major) do var p, i, j, k, base;
	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	j := 0;
	base := -1;
	for (i=0, Nkw) do
		p := @Buffer::((j+2)*NCOLS);
		for (k=0, 8) p::k := '\s';
		if (str.strtonum(Index[i*LOCATION+L_name], 10, 0) = major) do
			if (base = -1) base := i;
			t.memcopy(@p::8, Index[i*LOCATION+L_name],
				str.length(Index[i*LOCATION+L_name])+1);
			j := j+1;
		end
	end
	Size := j+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
	return base;
end


build_topindex() do var p, i, j;
	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	for (i=0, Ntkw) do
		p := @Buffer::((i+2)*NCOLS);
		for (j=0, 8) p::j := '\s';
		t.memcopy(@p::8, Topindex[i*LOCATION+L_name],
			str.length(Topindex[i*LOCATION+L_name])+1);
	end
	Size := Ntkw+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
end


getk() do var k;
	k := tty.readc();
	Message := 0;
	return k;
end


puts(y, off, s) do var b::NCOLS;
	tty.move(0, y);
	tty.clreol();
	if (str.length(s) >= off) do
		ie (Cols < NCOLS) do
			t.memcopy(b, @s::off, NCOLS);
			b::Cols := 0;
			tty.writes(b);
		end
		else do
			tty.writes(@s::off);
		end
	end
end


findright() do
	var	max, i, j, s;

	max := 0;
	for (i=0, Size) do
		s := @Buffer::(i*NCOLS);
		j := str.length(s);
		if (j > max) max := j;
	end
	return max < Cols-> Cols: max;
end


wh(y, s) do
	tty.move(4, y);
	tty.writes(s);
end

help() do
	tty.color(C_text);
	tty.clear();
	msg(packed"Help screen -- press [ENTER] to return", 1);
	tty.move(0, 0);
	tty.clreol();
	tty.writes(packed"    THelp 1.0 Copyright (C) 1999, 2000 Nils M Holm");
	tty.color(C_text);
	wh(3,
	packed"[j][k]      Move the cursor up/down         [Up][Down]");
	wh(4,
	packed"[^F][^B]    Move to the next/prev. page     [Next][Prev]");
	wh(5,
	packed"[^A][^E]    Move to beginning/end of page   [Home][End]");
	wh(6,
	packed"[l]         Select current item             [Right][CR]");
	wh(7,
	packed"[h]         Return to parent / exit on top  [Left][ESC]");
	wh(8,
	packed"[+][-]      Move to next/prev. section      [ and ]");
	wh(9,
	packed"[Q]         Quit immediately                [Z]");
	wh(10,
	packed"[ ]         Move screen left/right");
	wh(11,
	packed"[^L]        Screen refresh");
	wh(12,
	packed"[K]         Keyword search (only in single-file mode)");
	wh(14,
	packed"For conditions of use, see the file LICENSE which is part");
	wh(15,
	packed"of the T3X distribution. ('tx6 -V' prints a copy.)");
	getk();
end


prompt(ps, buf) do
	var	dfl::11, k;
	var	cp, x;

	k := str.length(buf);
	ie (k > 7) do
		t.memcopy(dfl, buf, 7);
		str.copy(@dfl::7, packed"...");
	end
	else do
		str.copy(dfl, buf);
	end
	tty.move(0, Lines-1);
	tty.color(C_high);
	tty.clreol();
	tty.writes(ps);
	tty.writes(packed" [");
	tty.writes(dfl);
	tty.writes(packed"]:");
	cp := 0;
	x := str.length(ps)+str.length(dfl)+5;
	while (1) do
		tty.move(x+cp, Lines-1);
		k := tty.readc();
		ie (k = '\b' /\ cp) do
			cp := cp-1;
			tty.move(x+cp, Lines-1);
			tty.writec('\s');
		end
		else ie (k = '\r' \/ k = '\n') do
			if (\dfl::0) buf::cp := 0;
			return;
		end
		else ie (k = ttyctl.K_ESC \/ k = 'G'-'@') do
			buf::0 := 0;
			return -1;
		end
		else ie (k = 'U'-'@') do
			cp := 0;
			tty.move(0, Lines-1);
			tty.clreol();
			tty.writes(ps);
			tty.writes(packed" [");
			tty.writes(dfl);
			tty.writes(packed"]:");
		end
		else if (k >= '\s' /\ k <= '~' /\ cp+x < Cols-1) do
			tty.writec(k);
			buf::cp := k;
			cp := cp+1;
		end
		dfl::0 := 0;
	end
end


browse(base, slct, page) do
	var	curr, top, redraw, otop;
	var	k, i;
	var	offset;
	var	new;

	curr := base;
	top := base-Lines+2;
	if (top < 0) top := 0;
	tty.color(C_text);
	tty.clear();
	offset := 0;
	redraw := 1;
	while (1) do
		showtitle(top+Lines-1);
		if (redraw) do
			tty.color(C_text);
			for (i=0, Lines-1) do
				ie (i+top < Size) puts(i, offset,
					@Buffer::((top+i)*NCOLS));
				else
					puts(i, 0, packed"");
			end
			redraw := 0;
		end
		ie (slct) do
			tty.color(C_high);
			puts(curr-top, offset, @Buffer::(curr*NCOLS));
			ie (slct) tty.move(8, curr-top);
			else tty.move(0, Lines-1);
			k := getk();
			tty.color(C_text);
			puts(curr-top, offset, @Buffer::(curr*NCOLS));
		end
		else do
			ie (slct) tty.move(8, curr-top);
			else tty.move(0, Lines-1);
			k := getk();
			ie (k = ttyctl.K_DOWN \/ k = 'j') do
				curr := top+Lines-2;
				if (curr >= Size) curr := Size-1;
			end
			else if (k = ttyctl.K_UP \/ k = 'k') do
				curr := top;
			end
		end
		ie (	k = ttyctl.K_ESC \/ k = ttyctl.K_LEFT \/
			k = 'h' \/ k = 'q'
		) do
			return %1;
		end
		else ie (k = 'Q' \/ k = 'Z') do
			fini();
			halt 0;
		end
		else ie (k = ttyctl.K_DOWN \/ k = 'j') do
			if (curr >= Size-1) do
				loop;
			end
			curr := curr+1;
			if (curr > top+Lines-2) do
				top := top+1;
				tty.scroll(0, Lines-1);
				if (\slct) do
					tty.color(C_text);
					puts(curr-top, offset,
						@Buffer::(curr*NCOLS));
				end
			end
		end
		else ie (k = ttyctl.K_UP \/ k = 'k') do
			if (curr < 1) do
				loop;
			end
			curr := curr-1;
			if (curr < top) do
				top := top-1;
				tty.rscroll(0, Lines-1);
				if (\slct) do
					tty.color(C_text);
					puts(curr-top, offset,
						@Buffer::(curr*NCOLS));
				end
			end
		end
		else ie (k = ttyctl.K_NEXT \/ k = 'f' \/ k = 'F'-'@') do
			otop := top;
			curr := curr + Lines-2;
			top := top + Lines-2;
			if (curr >= Size) do
				curr := Size-1;
				if (curr < 0) curr := 0;
			end
			if (top > Size-Lines+2) do
				top := Size-Lines+1;
				if (top < 0) top := 0;
			end
			redraw := otop \= top;
		end
		else ie (k = ttyctl.K_PREV \/ k = 'b' \/ k = 'B'-'@') do
			otop := top;
			curr := curr - Lines+2;
			top := top - Lines+2;
			if (curr < 0) curr := 0;
			if (top < 0) top := 0;
			redraw := otop \= top;
		end
		else ie (k = ttyctl.K_END \/ k = 'E'-'@' \/ k = 'G') do
			otop := top;
			curr := Size-1;
			if (curr < 0) curr := 0;
			top := Size-Lines+1;
			if (top < 0) top := 0;
			redraw := otop \= top;
		end
		else ie (k = ttyctl.K_HOME \/ k = 'A'-'@' \/ k = '1') do
			otop := top;
			top := 0;
			curr := 0;
			redraw := otop \= 0;
		end	
		else ie (k = ttyctl.K_CR \/ k = ttyctl.K_RIGHT \/ k = 'l') do
			if (\slct \/ Buffer::(curr*NCOLS) = 0) loop;
			return curr-2;
		end
		else ie (k = '[' \/ k = ']' \/ k = '+' \/ k = '-') do
			if (\page) loop;
			return k = '[' \/ k = '-'-> PREVTOPIC: NEXTTOPIC;
		end
		else ie (k = '\s') do
			if (Cols < NCOLS) do
				offset := offset=0-> findright()-Cols: 0;
				redraw := 1;
			end
		end
		else ie (k = 'K') do
			if (Nfiles > 1) do
				msg(packed
				"Multiple files -- search disabled", 1);
				loop;
			end
			if (prompt(packed"Search", Keyword) < 0) loop;
			return LIST_KEYWORDS;
		end
		else ie (k = 'L'-'@') do
			redraw := 1;
		end
		else ie (k = ttyctl.K_F1 \/ k = '?') do
			help();
			redraw := 1;
		end
		else do
			msg(packed"Press '?' for help", 1);
		end
	end
end


findleft(src) do
	var	l, i, j;

	l := 256;
	i := 0;
	while (src::i \/ src::(i+1)) do
		for (j=0, l)
			if (src::(i+j) \= '\t')
				leave;
		if (src::(i+j) /\ j < l) l := j;
		while (src::i) i := i+1;
		i := i+1;
	end
	return l;
end


istable(src) do var i;
	i := 0;
	while (src::i = '\t' \/ src::i = '\s') i := i+1;
	if (src::i = '|' /\ src::(i+1) = '\s') return 0;
	while (src::i) do
		if (src::i = '\t') return 1;
		i := i+1;
	end
	return 0;
end


insert(line, pos, chr, n) do
	var	i;

	for (i=NCOLS-1, pos+n-1, -1) line::i := line::(i-n);
	for (i=0, n) line::(i+pos) := chr;
end


insertvbar(line, pos) do
	var	i;

	for (i=NCOLS-1, pos+2, -1) line::i := line::(i-3);
	line::(pos+0) := '\s';
	line::(pos+1) := Lineset::0;
	line::(pos+2) := '\s';
end


inserthline(where, len) do
	var	i, p;

	for (i=NLINES-1, where-1, -1) do
		p := @Buffer::(i*NCOLS);
		t.memcopy(p, p-NCOLS, NCOLS);
	end
	for (i=0, len) p::i := Lineset::1;
	p::i := 0;
	return p;
end


implode(line, pos) do
	var	n, i;

	n := 0;
	while (line::(pos+n) = '\t') n := n+1;
	for (i=pos, NCOLS-n) line::i := line::(i+n);
end


format_table(from, pto) do
	var	i, to, s;
	var	z[MAXCOLS], j, k, b, nk, nb;
	var	ul, ll;

	to := pto[0];
	for (j=0, MAXCOLS) z[j] := 0;
	nk := 0;
	for (i=from, to) do
		s := @Buffer::(i*NCOLS);
		j := 0;
		k := 0;
		while (k < MAXCOLS /\ s::j) do
			while (s::j = '\t') j := j+1;
			b := j;
			while (s::j \= '\t' /\ s::j) j := j+1;
			if (j-b > z[k]) z[k] := j-b;
			k := k+1;
		end
		if (k > nk) nk := k;
	end
	for (i=from, to) do
		s := @Buffer::(i*NCOLS);
		j := 0;
		k := 0;
		while (s::j = '\t') j := j+1;
		while (k < MAXCOLS /\ s::j) do
			b := j;
			while (s::j \= '\t' /\ s::j) j := j+1;
			if (j-b = 1 /\ s::b = '.') s::b := '\s';
			if (\s::j) leave;
			implode(s, j);
			nb := z[k]-j+b;
			insert(s, j, '\s', nb);
			j := j+nb;
			insertvbar(s, j);
			j := j + 3;
			k := k+1;
		end
		nb := z[k]-j+b;
		insert(s, j, '\s', nb);
	end
	k := 0;
	for (j=0, nk) k := k+z[j];
	k := k + (nk-1)*3;
	ul := inserthline(from, k);
	ll := inserthline(to+1, k);
	j := 0;
	for (i=0, nk-1) do
		j := j+z[i]+1;
		ul::j := Lineset::2;
		ll::j := Lineset::3;
		j := j+2;
	end
	to := to+2;
	s := @Buffer::(from*NCOLS);
	k := str.length(s);
	for (i=from, to) do
		s := @Buffer::(i*NCOLS);
		insert(s, k, '\s', 2);
		s::(k+1) := Lineset::0;
		insert(s, 0, '\s', 2);
		s::0 := Lineset::0;
	end
	s := @Buffer::(from*NCOLS);
	s::0 := Lineset::4;
	s::1 := Lineset::1;
	s::(k+3) := Lineset::5;
	s::(k+2) := Lineset::1;
	s := @Buffer::((to-1)*NCOLS);
	s::0 := Lineset::6;
	s::1 := Lineset::1;
	s::(k+3) := Lineset::7;
	s::(k+2) := Lineset::1;
	pto[0] := to;
end


format_topic(isrc) do
	var	left, line, i, k, dest;
	var	intbl;
	var	toolong;
	var	src;

	clear();
	msg(packed"Formatting...", 0);
	setlocation(isrc[L_hseg], isrc[L_lseg]);
	src := @File::(isrc[L_offset]);
	left := findleft(src);
	Buffer::0 := 0;
	dest := 1;
	intbl := -1;
	i := 0;
	toolong := 0;
	while (src::i \/ src::(i+1)) do
		if (dest >= NLINES) do
			toolong := 1;
			leave;
		end
		ie (src::i = 0) do
			Buffer::(dest*NCOLS) := 0;
			line := packed"";
		end
		else do
			line := @src::(i+left);
			k := str.length(line);
			if (line::0 = '|') do
				ie (line::1 = '\s') line := line+2;
				else if (\line::1) line := line+1;
			end
			t.memcopy(@Buffer::(dest*NCOLS), line, k+1);
			i := i + k+left;
		end
		ie (istable(line)) do
			if (intbl = -1) intbl := dest;
		end
		else do
			if (intbl \= -1) format_table(intbl, @dest);
			intbl := -1;
		end
		dest := dest+1;
		i := i+1;
	end
	if (intbl \= -1) do
		if (dest >= NLINES-2) do
			toolong := 1;
			dest := dest-2;
		end
		format_table(intbl, @dest);
	end
	Buffer::(dest*NCOLS) := 0;
	Size := dest+1;
	ie (toolong)
		msg(packed"Topic too long!", 1);
	else
		showtitle(-1);
end


upcase(a, b) do var i;
	i := 0;
	while (a::i) do
		b::i := 'a' <= a::i /\ a::i <= 'z'-> a::i-'a'+'A': a::i;
		i := i+1;
	end
	b::i := 0;
end


build_keylist() do
	var p, i, j, top, kbuf::128, tbuf::128;

	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	upcase(Keyword, kbuf);
	j := 0;
	top := 0;
	for (i=0, Nkw) do
		upcase(Index[i*LOCATION+L_name], tbuf);
		if (str.find(tbuf, kbuf) \= -1) do
			ie (j >= NLINES-4) do
				msg(packed"Too many matches", 1);
				leave;
			end
			else do
				t.memfill(p, '\s', 8);
				str.copy(@p::8, tbuf);
				p := p+NCOLS;
				Keyndx[j] := i;
				j := j+1;
			end
		end
	end
	Size := j+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
end


viewfile(name) do var k, mode, sect_off, topic_off, sect_len, base;
	loadfile(name);
	mode := Keyword::0-> M_KEY: M_TOP;
	Section := 0;
	Topic := 0;
	sect_off := 0;
	topic_off := 0;
	while (1) do
		ie (mode = M_TOP) do
			build_topindex();
			k := browse(sect_off+2, 1, 0);
		end
		else ie (mode = M_INDEX) do
			base := build_subindex(str.strtonum(Section, 10, 0));
			ie (Size >= 5) do
				k := browse(topic_off+2, 1, 0);
				sect_len := Size-4;
			end
			else do
				format_topic(@Topindex[sect_off*LOCATION]);
				k := browse(0, 0, 1);
				sect_len := 0;
			end
		end
		else ie (mode = M_VIEW) do
			format_topic(@Index[(base+topic_off)*LOCATION]);
			k := browse(0, 0, 1);
		end
		else do	! mode = M_KEY
			if (\Keyword::0) do
				mode := M_TOP;
				loop;
			end
			build_keylist();
			if (Size < 5) do
				msg(packed"Keyword not found", 1);
				mode := M_TOP;
				loop;
			end
			k := Size=5-> 0: browse(2, 1, 0);
			ie (k < 0) do
				mode := M_TOP;
			end
			else do
				k := Keyndx[k];
				sect_off := str.strtonum(
					Index[k*LOCATION+L_name], 10, 0)-1;
				base := build_subindex(sect_off+1);
				topic_off := k-base;
				Topic := Index[(base+topic_off)*
					LOCATION+L_name];
				Section := Topindex[sect_off*LOCATION+L_name];
				mode := M_VIEW;
			end
			loop;
		end
		ie (k = -1) do
			mode := mode-1;
			if (mode < M_VIEW) Topic := 0;
			if (mode < M_INDEX) Section := 0;
			if (mode < M_TOP) leave;
		end
		else ie (k = PREVTOPIC) do
			ie (topic_off > 0) do
				topic_off := topic_off - 1;
				Topic := Index[(base+topic_off)*
					LOCATION+L_name];
			end
			else
				msg(packed"Beginning of section", 1);
		end
		else ie (k = NEXTTOPIC) do
			ie (topic_off < sect_len-1) do
				topic_off := topic_off + 1;
				Topic := Index[(base+topic_off)*
					LOCATION+L_name];
			end
			else
				msg(packed"End of section", 1);
		end
		else ie (k = LIST_KEYWORDS) do
			mode := M_KEY;
		end
		else do
			ie (mode = M_TOP) do
				Section := Topindex[k*LOCATION+L_name];
				sect_off := k;
				topic_off := 0;
			end
			else do
				Topic := Index[(base+k)*LOCATION+L_name];
				topic_off := k;
			end
			mode := mode+1;
		end
	end
end


getfileargs(ap) do var abuf::128, k;
	while (1) do
		k := t.getarg(ap, abuf, 128);
		if (k < 0) leave;
		if (Nfiles >= MAXFILES) fatal(packed"Too many files", 0);
		Filelist[Nfiles] := savename(abuf);
		Nfiles := Nfiles+1;
		ap := ap+1;
	end
end


read_sdv_config() do
	object	s[iostream];
	var	sbuf::SBUFLEN;
	var	lbuf::128;
	var	k, i, j;

	if (s.open(packed"sdv-config", sbuf, SBUFLEN,
		iostream.FREAD|iostream.FKILLCR) \= -1
	) do
		while (1) do
			k := s.reads(lbuf, 128);
			if (s.eof()) leave;
			if (lbuf::(k-1) = '\n') lbuf::(k-1) := 0;
			if (\t.memcomp(lbuf, packed"file", 4)) do
				i := 4;
				while (	lbuf::i = '\t' \/ lbuf::i = '\s' \/
					lbuf::i = '='
				)
					i := i+1;
				j := i;
				while (lbuf::i \= '\s' /\ lbuf::i)
					i := i+1;
				if (lbuf::i = '\s') lbuf::i := 0;
				if (Nfiles >= MAXFILES)
					fatal(packed"Too many files", 0);
				Filelist[Nfiles] := savename(@lbuf::j);
				Nfiles := Nfiles+1;
			end
		end
		s.close();
	end
end


scan_sd_files() do
	var	dd, k, buf::128;

	dd := sys.opendir(packed".");
	while (1) do
		k := sys.readdir(dd, buf, 128);
		if (k > 3 /\ (\str.comp(@buf::(k-3), packed".sd") \/
			\str.comp(@buf::(k-3), packed".SD"))
		) do
			if (Nfiles >= MAXFILES)
				fatal(packed"Too many files", 0);
			Filelist[Nfiles] := savename(buf);
			Nfiles := Nfiles+1;
		end
		if (k < 0) leave;
	end
	sys.closedir(dd);
end


getname(fname, name) do var buf::256, fd, i, j;
	fd := t.open(fname, T3X.OREAD);
	if (fd < 0) fatal(packed"Cannot open file", fname);
	t.read(fd, buf, 256);
	t.close(fd);
	i := 0;
	while (	(buf::i = '\s' \/ buf::i = '\t' \/ buf::i = '\n') /\
		i < 256
	)
		i := i+1;
	j := i;
	while (buf::i \= '\n' /\ i < 256) i := i+1;
	buf::i := 0;
	if (i-j >= NCOLS-16) buf::(j+NCOLS-17) := 0;
	str.copy(name, @buf::j);
end


build_filelist() do
	var	p, i;
	var	name::NCOLS;

	clear();
	p := @Buffer::(NCOLS*2);
	Buffer::0 := 0;
	Buffer::NCOLS := 0;
	for (i=0, Nfiles) do
		getname(Filelist[i], name);
		t.memfill(p, '\s', 8);
		str.copy(@p::8, name);
		p := p+NCOLS;
	end
	Size := Nfiles+4;
	Buffer::((Size-1)*NCOLS) := 0;
	Buffer::((Size-2)*NCOLS) := 0;
	str.copy(Title, packed"Select file");
end


makefilelist(fileargs) do
	if (\Nfiles) getfileargs(fileargs);
	if (\Nfiles) read_sdv_config();
	if (\Nfiles) scan_sd_files();
	Nbase := Nt;
end


writes(s) do var b::3;
	if (s) t.write(T3X.SYSOUT, s, str.length(s));
	t.write(T3X.SYSOUT, t.newline(b), str.length(b));
end


usage() do
	fini();
	writes(0);
	writes(packed"Usage: thelp [-l] [-m] [-s] [-c color] [-k keyword]");
	writes(packed"             [-s section] [file ...]");
	writes(0);
	writes(packed"-l      use nice-looking IBM line drawing characters");
	writes(packed"-m      monochrome (actually b&w) mode");
	writes(packed"-p      assume HP1000 palmtop (64x18) screen");
	writes(packed"-c XX   use color code XX (where X is a hex digit)");
	writes(packed"-d dir  search files in the given directory");
	writes(packed"-k name search the given keyword (in single file)");
	writes(packed"-s sect select a specific section: C=classes,");
	writes(packed"        T=T3X compiler, U=utilities, F=file formats,");
	writes(packed"        8=8086 assembler");
	writes(packed"file... file(s) to display");
	writes(0);
end


xtoi(s) do var c0, c1;
	c0 := s::0;
	c1 := s::1;
	c0 := '0'<=c0/\c0<='9'-> c0-'0': 'a'<=c0/\c0<='f'-> c0-'a'+10:
		'A'<=c0/\c0<='F'-> c0-'A'+10: 0;
	if (\c1) return c0;
	c1 := '0'<=c1/\c1<='9'-> c1-'0': 'a'<=c1/\c1<='f'-> c1-'a'+10:
		'A'<=c1/\c1<='F'-> c1-'A'+10: 0;
	return c0<<4|c1;
end


selectsect(id) do
	Nfiles := 1;
	ie (id::0='c' \/ id::0='C') Filelist[0] := packed"classes.sd";
	else ie (id::0='t' \/ id::0='T') Filelist[0] := packed"compiler.sd";
	else ie (id::0='u' \/ id::0='U') Filelist[0] := packed"utilities.sd";
	else ie (id::0='f' \/ id::0='F') Filelist[0] := packed"formats.sd";
	else ie (id::0='8') Filelist[0] := packed"s86.sd";
	else Nfiles := 0;
end


argarg(abuf, buf, pap, pi, plen) do
	ie (abuf::(pi[0]+1)) do
		str.copy(buf, @abuf::(pi[0]+1));
	end
	else do
		pap[0] := pap[0]+1;
		plen[0] := t.getarg(pap[0], abuf, 128);
		if (plen[0] < 1) do
			usage();
			halt;
		end
		str.copy(buf, abuf);
	end
	pi[0] := plen[0]-1;
end


config() do var	abuf::128, aabuf::128, ap, i, len;
	ap := 1;
	while (1) do
		len := t.getarg(ap, abuf, 128);
		if (len < 1 \/ abuf::0 \= '-') leave;
		i := 1;
		while (abuf::i) do
			ie (abuf::i = 'm') do
				C_text := 0x07;
				C_high := 0x70;
			end
			else ie (abuf::i = 'l') do
				Lineset := packed[ 179, 196, 194, 193, 218,
				191, 192, 217, 177, 176 ];
			end
			else ie (abuf::i = 'p') do
				Lines := 18;
				Cols := 64;
			end
			else ie (abuf::i = 'c') do
				argarg(abuf, aabuf, @ap, @i, @len);
				C_text := xtoi(aabuf);
				C_high := (C_text>>4) | (C_text<<4&0xff)
					& 0xF7;
			end
			else ie (abuf::i = 'd') do
				argarg(abuf, Directory, @ap, @i, @len);
			end
			else ie (abuf::i = 'k') do
				argarg(abuf, Keyword, @ap, @i, @len);
			end
			else ie (abuf::i = 's') do
				argarg(abuf, aabuf, @ap, @i, @len);
				selectsect(aabuf);
			end
			else do
				usage();
				halt 1;
			end
			i := i+1;
		end
		ap := ap+1;
	end
	return ap;
end


do var k;
	init();
	k := config();
	sys.chdir(Directory);
	makefilelist(k);
	if (Nfiles < 1) fatal(packed"No files found", 0);
		ie (Nfiles > 1) do
			if (Keyword::0)
				msg(packed
				"Multiple files -- keyword ignored", 1);
			Keyword::0 := 0;
			k := 0;
			while (1) do
				build_filelist();
				k := browse(k+2, 1, 0);
				if (k < 0) leave;
				viewfile(Filelist[k]);
			end
		end
		else do
			viewfile(Filelist[0]);
		end
		fini();
end
