!	TX6 -- T3X-R6 compiler driver
!	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.
!
!	SYNTAX: TX [options] file ...

#classpath "../classes";

module tx(t3x, system, string, iostream);

object	t[t3x],
	sys[system],
	str[string];


struct	SYMREC = SNAME, SVAL, SSIZE, SFLAGS;	! From TXTRN

const	MAXARGLEN	= 128;
const	STRPOOLSIZE	= 4096;
const	MAXNAMES	= 128;
const	MAXCLASSES	= 128;
const	BUFLEN		= 512;


var	O_bindir,
	O_classdir,
	O_libdir,
	O_dpfilter,
	O_optimize,
	O_userconfig,
	O_verbose,
	O_compile_only,
	O_make_class,
	O_keep_tmpfiles,
	O_output_file,
	O_tmpin,
	O_tmpout,
	O_tmppfx,
	O_tmpdir,
	O_keep_cdir,
	O_r5_compat,
	O_bang_header,
	O_target,
	O_as_cmd,
	O_ld_cmd,
	O_ldctrl,
	O_exesfx,
	O_libs;

var	Tmp_hack;

var	Basedir;

var	Strpool::STRPOOLSIZE, St;
var	Names[MAXNAMES], Nt, Keeplist::MAXNAMES;
var	Classes[MAXCLASSES], Ct;


init() do
	sys.init();
	St := 0;
	Nt := 0;
	Ct := 0;
	Tmp_hack := 0;
end


wr(s) t.write(1, s, str.length(s));

ws(s) do var nl::3;
	t.write(1, s, str.length(s));
	wr(t.newline(nl));
end

wm(s) t.write(2, s, str.length(s));

wf(tp, a) do var buf::256;
	return ws(str.format(buf, tp, a));
end


remove(path, force) do
	if (O_keep_tmpfiles /\ \force) do
		if (O_verbose > 1) wf(packed"(keep) %s", [(path)]);
		return;
	end
	if (O_verbose > 1) wf(packed"(delete) %s", [(path)]);
	return t.remove(path);
end


error(m, n) do var nl::3;
	wm(packed"TX: ");
	wm(m);
	if (n) do
		wm(packed": ");
		wm(n);
	end
	wm(t.newline(nl));
	remove(O_tmpin, 0);
	remove(O_tmpout, 0);
	t.remove(packed"T3X.ERR");
	halt 1;
end


savestr(s) do var k;
	k := str.length(s)+1;
	if (k+St >= STRPOOLSIZE) error(packed"out of string space", 0);
	t.memcopy(@Strpool::St, s, k);
	St := St+k;
	return @Strpool::(St-k);
end


defaults() do var tmp::MAXARGLEN+1, tmp2::MAXARGLEN+10;
	ie (t.getenv(packed"T3XDIR", tmp, MAXARGLEN) < 0) do
		O_bindir := packed".";
		O_classdir := packed".";
		O_libdir := packed".";
		Basedir := packed".";
	end
	else do
		O_bindir := savestr(str.format(tmp2, packed"%S/bin", [(tmp)]));
		O_classdir := savestr(str.format(tmp2, packed"%S/classes",
			[(tmp)]));
		O_libdir := savestr(str.format(tmp2, packed"%S/lib", [(tmp)]));
		Basedir := savestr(tmp);
	end
	O_dpfilter := 1;
	O_optimize := 1;
	O_userconfig := packed"~/.txrc";
	O_verbose := 0;
	O_compile_only := 0;
	O_make_class := 0;
	O_keep_tmpfiles := 0;
	O_output_file := 0;
	O_tmpin := 0;
	O_tmpout := 0;
	O_tmpdir := 0;
	O_keep_cdir := 0;
	O_r5_compat := 0;
	O_bang_header := 0;
	O_target := 0;
	O_as_cmd := packed"/usr/bin/as -o %S %S";
	O_ld_cmd := packed
	"/usr/bin/ld -Bforcedynamic -s -o %S /usr/lib/crt0.o %S";
	O_libs := packed"%S/libt3x.a /usr/lib/libc.a /usr/lib/libgcc.a";
	O_ldctrl := "";
	O_exesfx := 0;
end


savename(s, keep) do
	if (Nt >= MAXNAMES) error(packed"too many file names", 0);
	Names[Nt] := savestr(s);
	Keeplist::Nt := keep;
	Nt := Nt+1;
end


saveclass(s) do var i;
	for (i=0, Ct)
		if (\str.comp(Classes[i], s))
			return;
	if (Ct >= MAXCLASSES) error(packed"too many classes", 0);
	Classes[Ct] := savestr(s);
	Ct := Ct+1;
end


license() do
ws(packed
"T3X -- A Compiler for the Minimum Procedural Language T3X");
ws(packed
"Copyright (C) 1996-2000 Nils M Holm.  All rights reserved.");
ws(packed"");
ws(packed
"Redistribution and use in source and binary forms, with or without");
ws(packed
"modification, are permitted provided that the following conditions");
ws(packed
"are met:");
ws(packed
"1. Redistributions of source code must retain the above copyright");
ws(packed
"   notice, this list of conditions and the following disclaimer.");
ws(packed
"2. Redistributions in binary form must reproduce the above copyright");
ws(packed
"   notice, this list of conditions and the following disclaimer in the");
ws(packed
"   documentation and/or other materials provided with the distribution.");
ws(packed"");
ws(packed
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND");
ws(packed
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
ws(packed
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
ws(packed
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE");
ws(packed
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
ws(packed
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS");
ws(packed
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)");
ws(packed
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
ws(packed
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
ws(packed
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF");
ws(packed
"SUCH DAMAGE.");
ws(packed"");
halt 0;
end


usage(hlt) do
	wr(packed"Usage: tx [-V] [-h?] [-B dir] [-C dir] [-L dir]");
	ws(packed" [-T dir] [-A arch]");
	ws(packed"\t[-U file] [-m target] [-o file] [-5DOceikpv] file ...");
	if (hlt) halt 1;
end


longusage() do
	var	first, i, Targets, tp::MAXARGLEN*2;

	Targets := [
		packed"386", packed"386a", packed"386e",
		packed"86", packed"86s", packed"exe",
		0
	];
	ws(packed"");
	usage(0);
	ws(packed"");
	ws(packed"-A name use architecture setup 'name'");
	ws(packed"-B dir  search binaries in 'dir'");
	ws(packed"-C dir  search classes in 'dir'");
	ws(packed"-D{01}  enable/disable dead procedure filter");
	ws(packed"-L dir  search libraries in 'dir'");
	ws(packed"-O{01}  enable/disable tcode optimization");
	ws(packed"-T dir  place tmp files in 'dir' (default: $TMP)");
	ws(packed"-U file load user config from 'file'");
	ws(packed"-V      display the conditions of use");
	ws(packed"-c      compile only (don't load)");
	ws(packed"-e      create a #! header (Unix only)");
	ws(packed"-h      display this summary (also -?)");
	ws(packed"-i      create installable class files");
	ws(packed"-k      keep intermediate files");
	ws(packed"-m type generate native code for target 'type'");
	ws(packed"-o file direct output to 'file'");
	ws(packed"-p      preserve CLASSDIR");
	ws(packed"-v      verbose operation (more v's = more output)");
	ws(packed"-5      enable T3X-R5 compatibility mode");
	ws(packed"");
	wr(packed"Installed targets: ");
	first := 1;
	i := 0; while (Targets[i]) do
		str.format(tp, packed"%S/txcg%S%S", [(O_bindir),
			(Targets[i]), (O_exesfx-> O_exesfx: packed"")]);
		if (t.close(t.open(tp, T3X.OREAD)) = 0) do
			if (\first) wr(packed", ");
			first := 0;
			wr(Targets[i]);
		end
		i := i+1;
	end
	ws(packed"");
	ws(packed"");
	halt 0;
end


flag(s, i, p) do var c;
	c := s::(i+1);
	if (c = '0') do
		p[0] := 0;
		return i+1;
	end
	if (c = '1') do
		p[0] := 1;
		return i+1;
	end
	p[0] := 1;
	return i;
end


file(pi, s, pj, pk, p) do var i, j, k;
	i := pi[0];
	j := pj[0];
	k := pk[0];
	ie (s::(j+1)) do
		j := j+1;
	end
	else do
		i := i+1;
		k := t.getarg(i, s, MAXARGLEN);
		if (k < 0) usage(1);
		j := 0;
	end
	p[0] := savestr(@s::j);
	pk[0] := k;
	pj[0] := str.length(s)-1;
	pi[0] := i;
end


yesno(x) return x-> packed"YES": packed"NO";


printargs() do
	wf(packed"verbosity level = %D", [(O_verbose)]);
	wf(packed"userconfig = %S", [(O_userconfig)]);
	wf(packed"bindir = %S", [(O_bindir)]);
	wf(packed"classdir = %S", [(O_classdir)]);
	wf(packed"libdir = %S", [(O_libdir)]);
	wf(packed"exe suffix = %S", [(O_exesfx-> O_exesfx: packed"(none)")]);
	wf(packed"output file = %S", [(O_output_file-> O_output_file:
		packed"(automatic)")]);
	wf(packed"tmp prefix = %S", [(O_tmppfx)]);
	wf(packed"dead procedure filter enabled = %S",
		[(yesno(O_dpfilter))]);
	wf(packed"optimization enabled = %S", [(yesno(O_optimize))]);
	wf(packed"target = %S", [(O_target-> O_target: packed"(none)")]);
	wf(packed"compile only = %S", [(yesno(O_compile_only))]);
	wf(packed"keep class directory = %S", [(yesno(O_keep_cdir))]);
	wf(packed"keep tmpfiles = %S", [(yesno(O_keep_tmpfiles))]);
	wf(packed"release 5 compatibility = %S", [(yesno(O_r5_compat))]);
	wf(packed"assembler command = %S", [(O_as_cmd)]);
	wf(packed"loader command = %S", [(O_ld_cmd)]);
	if (O_verbose > 3) wf(packed"loader control file = %S", [(O_ldctrl)]);
	wf(packed"libraries = %S", [(O_libs)]);
end


source_p(s) do var i;
	i := str.rscan(s, '.');
	if (i < 0) return 0;
	return str.comp(@s::i, packed".t") = 0;
end


exists(name) return t.close(t.open(name, T3X.OREAD)) = 0;


copy(in, out, outname) do
	const	CPBUFLEN = 1024;
	var	cpbuf::CPBUFLEN;
	var	k;

	k := t.read(in, cpbuf, CPBUFLEN);
	while (k > 0) do
		if (t.write(out, cpbuf, k) \= k)
			error(packed"file write error", outname);
		k := t.read(in, cpbuf, CPBUFLEN);
	end
end


rename(old, new, cp) do var buf::MAXARGLEN*2+5, in, out;
	if (O_verbose > 1) wf(packed"(rename) %s -> %s", [(old), (new)]);
	if (t.rename(old, new) < 0) do
		ie (cp) do
			if (O_verbose > 1) wf(packed
				"(copying instead) %s -> %s", [(old), (new)]);
			in := t.open(old, T3X.OREAD);
			if (in < 0) error(packed"could not open", old);
			out := t.open(new, T3X.OWRITE);
			if (out < 0) error(packed"caould not create", old);
			copy(in, out, new);
			t.close(in);
			t.close(out);
			t.remove(old);
		end
		else do
			str.format(buf, packed"%S -> %S", [(old), (new)]);
			error(packed"failed to rename", buf);
		end
	end
end


run(cmd) do
	const	NARG = 18;
	var	args[NARG];
	var	p, i;

	i := 0;
	p := 0;
	while (1) do
		while (cmd::i = '\s' \/ cmd::i = '\t')
			i := i+1;
		if (\cmd::i) leave;
		if (p >= NARG-1) error(packed"too many aruments", cmd);
		args[p] := @cmd::i; p := p+1;
		while (cmd::i /\ cmd::i \= '\s' /\ cmd::i \= '\t')
			i := i+1;
		if (cmd::i) do
			cmd::i := 0;
			i := i+1;
		end
	end
	args[p] := 0;
	if (O_verbose) do var nl::3;
		for (i=0, p) do
			if (i) wr("\s");
			wr(args[i]);
		end
		wr(t.newline(nl));
	end
	if (sys.spawn(args[0], @args[1], system.SPAWN_WAIT))
		error(packed"command returned failure", args[0]);
end


run_with_redirection(prog, opts, infile, outfile, append) do
	var	old0, old1, in, out;
	var	trnpath::MAXARGLEN+6;
	var	rc, k, i;

	ie (prog::0 = '!') do
		str.copy(trnpath, @prog::1);
	end
	else ie (prog::0 = '/') do
		ie (O_exesfx)
			str.format(trnpath, packed"%S%S",
				[(prog), (O_exesfx)]);
		else
			str.copy(trnpath, prog);
	end
	else do
		str.format(trnpath, packed"%S/%S%S",
			[(O_bindir), (prog), (O_exesfx-> O_exesfx: "")]);
	end
	if (O_verbose) do
		wr(trnpath);
		i := 0; while (opts[i]) do
			wr(packed"\s");
			wr(opts[i]);
			i := i+1;
		end
		wr(packed": ");
		wf(packed"%S -> %S", [(infile),
			(outfile-> outfile: packed"(null)")]);
	end
	in := t.open(infile, T3X.OREAD);
	if (in < 0) error(packed"cannot open", infile);
	ie (append) do
		out := t.open(outfile, T3X.ORDWR);
		if (out < 0 \/ t.seek(out, 0, T3X.SEEK_END) < 0) do
			t.close(in);
			error(packed"cannot append", outfile);
		end
	end
	else if (outfile) do
		out := t.open(outfile, T3X.OWRITE);
		if (out < 0) do
			t.close(in);
			error(packed"cannot create", outfile);
		end
	end
	old0 := sys.dup(0);
	if (outfile) old1 := sys.dup(1);
	if (old0 < 0 \/ sys.dup2(in, 0) < 0)
		error(packed"redirection failed (this should not happen)", 0);
	if (outfile /\ (old1 < 0 \/ sys.dup2(out, 1) < 0))
		error(packed"redirection failed (this should not happen)", 0);
	rc := sys.spawn(trnpath, opts, system.SPAWN_WAIT);
	t.close(in);
	if (outfile) t.close(out);
	if (sys.dup2(old0, 0) < 0 \/ t.close(old0) < 0)
		error(packed"redirection cleanup failed (oops!)", 0);
	if (outfile /\ (sys.dup2(old1, 1) < 0 \/ t.close(old1) < 0))
		error(packed"redirection cleanup failed (oops!)", 0);
	if (rc = 255) error(packed"program not found", trnpath);
	if (exists(packed"T3X.ERR") \/ rc) do
		k := str.length(infile);
		if (k > 4 /\ \str.comp(@infile::(k-4), packed".tmp"))
			remove(infile, 0);
		if (outfile) remove(outfile, 0);
		error(packed"error in compilation -- terminating", 0);
	end
end


basename(a, b) do var i;
	str.copy(b, a);
	i := str.rscan(b, '.');
	if (i = %1) error(packed"file name has no suffix", a);
	b::i := 0;
	return b;
end


compile(s) do
	var	buf::MAXARGLEN+11, tname::MAXARGLEN*2+6;
	var	i;

	basename(s, buf);
	i := str.rscan(buf, '/');
	str.format(tname, packed"%S/%S.tcode", [(O_tmpdir), (@buf::(i+1))]);
	run_with_redirection(packed"txtrn", [0], s, tname, 0);
	savename(tname, 0);
end


copyclasses(outfile) do
	var	modname::MAXARGLEN*2+1, i;
	var	mf;

	for (i=0, Ct) do
		if (\str.comp(Classes[i], packed"t3x")) loop;
		str.format(modname, packed"./%S.tcode", [(Classes[i])]);
		mf := t.open(modname, T3X.OREAD);
		if (mf < 0) do
			str.format(modname, packed"%S/%S.cls",
				[(O_classdir), (Classes[i])]);
			mf := t.open(modname, T3X.OREAD);
			if (mf < 0) error(packed"cannot open class", modname);
		end
		if (O_verbose > 1) wf(packed"(load) %S", [(modname)]);
		copy(mf, outfile, modname);
		t.close(mf);
	end
end


resolve(clss) do
	object	ndx[iostream];
	var	nbuf::BUFLEN;
	var	path::MAXARGLEN*2;
	var	line::MAXARGLEN+1, name::MAXARGLEN+1;
	var	k, j;

	str.format(path, packed"%S/CLASSNDX", [(O_classdir)]);
	if (ndx.open(path, nbuf, BUFLEN, iostream.FREAD) = %1)
		return;
	k := ndx.reads(line, MAXARGLEN);
	while (\ndx.eof()) do
		line::(k-1) := 0;
		if (line::(k-2) = '\r') line::(k-2) := 0;
		j := str.length(clss);
		if (\t.memcomp(clss, line, j) /\ line::j = ':') do
			while (	str.parse(@line::j, packed"%[:,]%S,",
				[(name)]) = 1
			) do
				j := j+str.length(name)+1;
				saveclass(name);
			end
		end
		k := ndx.reads(line, MAXARGLEN);
	end
	ndx.close();
end


loadclasses(outfile) do
	object	cdir[iostream];
	var	cdbuf::BUFLEN;
	var	magic::4;
	var	corrdir;
	var	nb::4, n;
	var	c;
	var	b;
	var	modname::MAXARGLEN*2+1, i, j, k, p;

	corrdir := packed"corrupt CLASSDIR";
	if (O_r5_compat) do
		saveclass(packed"string");
		saveclass(packed"r5c");
		copyclasses(outfile);
		return;
	end
	b := 2;
	if (cdir.open(packed"CLASSDIR", cdbuf, BUFLEN, iostream.FREAD) = %1)
		return;
	if (cdir.read(magic, 4) \= 4) error(corrdir, 0);
	if (t.memcomp(magic, packed"CDIR", 4)) error(corrdir, 0);
	if (cdir.read(nb, b) \= b) error(corrdir, 0);
	n := nb::1 << 8 | nb::0;
	if (cdir.move(n.*b.*SYMREC+8, iostream.SEEK_SET) < 0)
		error(corrdir, 0);
	c := cdir.rdch();
	if (cdir.eof()) error(corrdir, 0);
	while (\cdir.eof()) do
		if (c = '*') do
			c := cdir.rdch();
			i := 0;
			while (c /\ c \= '*' /\ \cdir.eof()) do
				if (i >= MAXARGLEN) do
					modname::i := 0;
					error(packed"module name too long",
						modname);
				end
				modname::i := c; i := i+1;
				c := cdir.rdch();
			end
			modname::i := 0;
			for (j=0, Nt) do
				p := Names[j];
				p := @p::(str.rscan(p, '/')+1);
				if (	\t.memcomp(p, modname, i) /\
					p::i = '.'
				)
					leave;
			end
			if (j<Nt) leave;
			resolve(modname);
			if (j >= Nt) saveclass(modname);
		end
		c := cdir.rdch();
	end
	cdir.close();
	copyclasses(outfile);
end


loadall() do
	var	i;
	var	outfile;
	var	tmpname::MAXARGLEN+5, tmpfile;
	var	xhdr::MAXARGLEN+1;
	var	infile;
	var	ofile;

	outfile := t.open(O_tmpin, T3X.OWRITE);
	if (outfile < 0)
		error(packed"cannot create loader input file", O_tmpin);
	loadclasses(outfile);
	for (i=0, Nt) do
		infile := t.open(Names[i], T3X.OREAD);
		if (infile < 0) error(packed"cannot reopen", Names[i]);
		if (O_verbose > 1) wf(packed"(load) %S", [(Names[i])]);
		copy(infile, outfile, O_tmpin);
		t.close(infile);
		if (\Keeplist::i) remove(Names[i], 0);
	end
	t.close(outfile);
	if (O_bang_header) do
		ofile := t.open(O_tmpout, T3X.OWRITE);
		if (ofile < 0) error(packed"cannot create", O_tmpout);
		str.format(xhdr, packed"#! %S/txx\n", [(O_bindir)]);
		t.write(ofile, xhdr, str.length(xhdr));
		t.close(ofile);
	end
	run_with_redirection(packed"txld", [0], O_tmpin, O_tmpout,
		O_bang_header);
	remove(O_tmpin, 0);
end


optimize(prog, opts) do
	remove(O_tmpin, 1);
	rename(O_tmpout, O_tmpin, 0);
	run_with_redirection(prog, opts, O_tmpin, O_tmpout, 0);
	remove(O_tmpin, 0);
end


tcode_to_native(m) do
	var	cgprog::MAXARGLEN+1,
		s_path::MAXARGLEN+1,
		o_path::MAXARGLEN+1,
		tmp::MAXARGLEN+1,
		command::MAXARGLEN*2+1;
	var	k, ctlfd;

	str.format(cgprog, packed"%S/txcg%S%S",
		[(O_bindir), (m), (O_exesfx-> O_exesfx: "")]);
	if (t.close(t.open(cgprog, 0)) < 0)
		error(packed"unsupported architecture", m);
	str.format(cgprog, packed"txcg%S", [(m)]);
	if (	\str.comp(O_as_cmd, packed"(none)") /\
		\str.comp(O_ld_cmd, packed"(none)")
	) do
		str.format(tmp, packed"%S%S", [(O_output_file),
			(\str.comp(m, packed"exe")-> packed".exe": "")]);
		run_with_redirection(cgprog, Tmp_hack-> Tmp_hack: [0],
			O_tmpout, tmp, 0);
		remove(O_tmpout, 0);
		return;
	end
	str.format(s_path, packed"%S.s", [(O_tmppfx)]);
	run_with_redirection(cgprog, [0], O_tmpout, s_path, 0);
	remove(O_tmpout, 0);
	str.format(o_path, packed"%S.o", [(O_tmppfx)]);
	ie (str.comp(O_as_cmd, packed"(none)")) do
		ie (str.find(O_as_cmd, packed"%S") \= -1) do
			str.format(command, O_as_cmd, [(o_path), (s_path)]);
			run(command);
		end
		else do
			run_with_redirection(O_as_cmd, [0], s_path, o_path, 0);
		end
		remove(s_path, 0);
	end
	else do
		rename(s_path, o_path, 0);
	end
	ie (str.comp(O_ld_cmd, packed"(none)")) do
		str.format(tmp, O_libs, [(O_libdir)]);
		ie (str.find(O_ld_cmd, packed"%S") \= -1) do
			str.format(command, O_ld_cmd,
				[(O_output_file), (o_path)]);
			k := str.length(command);
			command::k := '\s';
			str.copy(@command::(k+1), tmp);
			run(command);
		end
		else do
			str.format(command, O_ldctrl,
				[(O_output_file), (o_path)]);
			k := str.length(command);
			str.copy(@command::k, tmp);
			ctlfd := t.open(packed"__ldctrl", T3X.OWRITE);
			if (ctlfd < 0)
				error(packed
					"failed to create __ldctrl file", 0);
			k := str.length(command);
			command::k := '\n'; k := k+1;
			if (t.write(ctlfd, command, k) \= k)
				error(packed
					"failed to write __ldctrl file", 0);
			t.close(ctlfd);
			run_with_redirection(O_ld_cmd, [0], packed"__ldctrl",
				0, 0);
			remove(packed"__ldctrl", 0);
		end
		remove(o_path, 0);
	end
	else do
		str.format(tmp, packed"%S%S", [(O_output_file),
			(O_exesfx-> O_exesfx: "")]);
		rename(o_path, tmp, 1);
	end
end


make_class() do
	var	cmd::MAXARGLEN+1, base::MAXARGLEN+1, cl_path::MAXARGLEN+1;

	str.format(cl_path, packed"%S.cls", [(O_output_file)]);
	remove(cl_path, 1);
	rename(Names[Nt-1], cl_path, 1);
	str.format(cmd, packed"%S/txcm%S -qfx %S CLASSDIR", [(O_bindir),
			(O_exesfx-> O_exesfx: packed""), (O_output_file)]);
	run(cmd);
	str.format(cmd, packed"%S/txcm%S -o %S.ndx -n %S CLASSDIR",
		[(O_bindir), (O_exesfx-> O_exesfx: packed""), (O_output_file),
		(O_output_file)]);
	run(cmd);
end


gettherest(buf, in, lim) do
	var	k, j, i, add::MAXARGLEN+1;

	k := str.length(buf);
	while (1) do
		if (buf::(k-1) = '\n') k := k-1;
		if (buf::(k-1) = '\r') k := k-1;
		if (buf::(k-1) \= '\\') leave;
		k := k-1;
		j := send(in, iostream, reads(add, MAXARGLEN));
		if (add::(j-1) = '\n') j := j-1;
		if (add::(j-1) = '\r') j := j-1;
		add::j := 0;
		for (i=0, j) if (add::i \= '\s' /\ add::i \= '\t') leave;
		if (j+k-i+1 >= lim)
			error(packed"config: string too long", 0);
		str.copy(@buf::k, @add::i);
		k := k+j-i;
	end
	return buf;
end


configure(path, reqd) do
	object	conf[iostream];
	var	buffer::BUFLEN;
	var	line::MAXARGLEN+1;
	var	tmp::MAXARGLEN*3+1;
	var	k, dummy;

	ie (path) do
		!
	end
	else ie (O_userconfig::0 = '~') do
		if (t.getenv(packed"HOME", tmp, MAXARGLEN) < 0)
			error(packed"cannot determine $HOME", 0);
		k := str.length(tmp);
		str.copy(@tmp::k, @O_userconfig::1);
		path := tmp;
	end
	else do
		path := O_userconfig;
	end
	if (O_verbose > 1) do
		wf(packed"(configure) %S", [(path)]);
	end
	if (conf.open(path, buffer, BUFLEN, iostream.FREAD) = %1) do
		if (reqd) error(packed"failed to open config file", path);
		if (\path) O_userconfig := packed"(none)";
		return;
	end
	k := conf.reads(line, MAXARGLEN);
	while (\conf.eof()) do
		line::(k-1) := 0;
		if (line::(k-2) = '\r') line::(k-2) := 0;
		if (\line::0 \/ line::0 = '#') do
			k := conf.reads(line, MAXARGLEN);
			loop;
		end
		ie (	line::0 = 'a' /\
			str.parse(line, packed"as_cmd%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_as_cmd := savestr(tmp);
		end
		else ie (line::0 = 'b' /\
			str.parse(line, packed"bindir%W=%W%S", [(@tmp)]) = 1
		)
			O_bindir := savestr(tmp);
		else ie (line::0 = 'c' /\
			str.parse(line, packed"classdir%W=%W%S", [(@tmp)]) = 1
		)
			O_classdir := savestr(tmp);
		else ie (line::0 = 'd' /\
			str.parse(line, packed"dpfilter%W=%W%D",
			[(@O_dpfilter)]) = 1
		)
			;
		else ie (line::0 = 'e' /\
			str.parse(line, packed"exesfx%W=%W%S", [(@tmp)]) = 1
		)
			O_exesfx := savestr(tmp);
		else ie (line::0 = 'f' /\
			str.parse(line, packed"force_cd4%W=%W%D",
			[(@dummy)]) = 1
		)
			error(packed"option is obsolete", packed"force_cd4");
		else ie (line::0 = 'l' /\
			str.parse(line, packed"libdir%W=%W%S", [(@tmp)]) = 1
		)
			O_libdir := savestr(tmp);
		else ie (line::0 = 'l' /\
			str.parse(line, packed"ld_cmd%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_ld_cmd := savestr(tmp);
		end
		else ie (line::0 = 'l' /\
			str.parse(line, packed"ldctrl%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_ldctrl := savestr(tmp);
			str.xlate(O_ldctrl, ';', '\n');
		end
		else ie (line::0 = 'l' /\
			str.parse(line, packed"libs%W=%W%S", [(@tmp)]) = 1
		) do
			gettherest(tmp, conf, MAXARGLEN*3);
			O_libs := savestr(tmp);
		end
		else ie (line::0 = 'o' /\
			str.parse(line, packed"optimize%W=%W%D",
			[(@O_optimize)]) = 1
		)
			;
		else ie (line::0 = 'r' /\
			str.parse(line, packed"r5compat%W=%W%D",
			[(@O_r5_compat)]) = 1
		)
			;
		else ie (line::0 = 't' /\
			str.parse(line, packed"target%W=%W%S", [(@tmp)]) = 1
		)
			O_target := savestr(tmp);
		else ie (line::0 = 't' /\
			str.parse(line, packed"tmpdir%W=%W%S", [(@tmp)]) = 1
		) do
			O_tmpdir := savestr(tmp);
			Tmp_hack := [ packed"-T", (O_tmpdir), 0 ];
		end
		else
			error(packed"error in user config", line);
		k := conf.reads(line, MAXARGLEN);
	end
	conf.close();
end


loadconf(id, reqd) do var buf::MAXARGLEN*2+1;
	str.format(buf, packed"%S/%S.txrc", [(Basedir), (id)]);
	if (reqd) do
		if (t.close(t.open(buf, T3X.OREAD)) < 0)
			error(packed"unsupported architecture", id);
	end
	configure(buf, reqd);
end


mktmpfiles() do
	var	buf::MAXARGLEN*2+17;
	var	tb::8;
	var	fd;

	ie (O_tmpdir)
		;
	else ie (t.getenv(packed"TMP", buf, MAXARGLEN+1) > 0)
		O_tmpdir := savestr(buf);
	else
		O_tmpdir := packed".";
	sys.time(tb);
	str.format(buf, packed"%S/tx%2:0D%2:0D%2:0D.in",
		[(O_tmpdir), (tb::4), (tb::5), (tb::6+tb::7)]);
	O_tmpin := savestr(buf);
	str.format(buf, packed"%S/tx%2:0D%2:0D%2:0D.out",
		[(O_tmpdir), (tb::4), (tb::5), (tb::6+tb::7)]);
	O_tmpout := savestr(buf);
	str.format(buf, packed"%S/tx%2:0D%2:0D%2:0D",
		[(O_tmpdir), (tb::4), (tb::5), (tb::6+tb::7)]);
	O_tmppfx := savestr(buf);
	fd := t.open(O_tmpin, T3X.OWRITE);
	ie (fd < 0) do
		error(packed"cannot create tmp file", O_tmpin);
	end
	else do
		t.close(fd);
		t.remove(O_tmpin);
	end
end


do
	var	abuf::MAXARGLEN*2+1, k, i, j;
	var	altarc;

	init();
	defaults();
	i := 1;
	loadconf(packed"system", 0);
	configure(0, 0);
	while (1) do
		k := t.getarg(i, abuf, MAXARGLEN);
		if (k < 0 \/ abuf::0 \= '-') leave;
		for (j=1, k) do
		ie (abuf::j = '4') error(packed"obsolete option", packed"-4");
		else ie (abuf::j = '5') O_r5_compat := 1;
		else ie (abuf::j = 'A') do
			file(@i, abuf, @j, @k, @altarc);
			loadconf(altarc, 1);
		end
		else ie (abuf::j = 'B') file(@i, abuf, @j, @k, @O_bindir);
		else ie (abuf::j = 'C') file(@i, abuf, @j, @k, @O_classdir);
		else ie (abuf::j = 'D') j := flag(abuf, j, @O_dpfilter);
		else ie (abuf::j = 'L') file(@i, abuf, @j, @k, @O_libdir);
		else ie (abuf::j = 'O') j := flag(abuf, j, @O_optimize);
		else ie (abuf::j = 'T') do
			file(@i, abuf, @j, @k, @O_tmpdir);
			Tmp_hack := [ packed"-T", (O_tmpdir), 0 ];
		end
		else ie (abuf::j = 'U') do
			file(@i, abuf, @j, @k, @O_userconfig);
			configure(0, 1);
		end
		else ie (abuf::j = 'V') license();
		else ie (abuf::j = 'c') O_compile_only := 1;
		else ie (abuf::j = 'e') O_bang_header := 1;
		else ie (abuf::j = 'h' \/ abuf::j = '?') longusage();
		else ie (abuf::j = 'i') O_make_class := 1;
		else ie (abuf::j = 'k') O_keep_tmpfiles := 1;
		else ie (abuf::j = 'm') file(@i, abuf, @j, @k, @O_target);
		else ie (abuf::j = 'o') file(@i, abuf, @j, @k, @O_output_file);
		else ie (abuf::j = 'p') O_keep_cdir := 1;
		else ie (abuf::j = 'v') O_verbose := O_verbose+1;
		else usage(1);
		end
		i := i+1;
	end
	mktmpfiles();
	if (O_verbose) ws(packed"T3X Release 6.8");
	if (O_verbose > 2) printargs();
	if (k < 0) usage(1);
	t.remove(packed"T3X.ERR");
	if (\O_keep_cdir) remove(packed"CLASSDIR", 1);
	while (1) do
		k := t.getarg(i, abuf, MAXARGLEN);
		if (k < 0) leave;
		ie (source_p(abuf)) do
			compile(abuf);
		end
		else do
			if (O_verbose)
				wf(packed"%S: no .t suffix (not compiling)",
					[(abuf)]);
			savename(abuf, 1);
		end
		i := i+1;
	end
	if (\O_output_file) do
		basename(Names[Nt-1], abuf);
		i := str.rscan(abuf, '/');
		O_output_file := savestr(@abuf::(i+1));
	end
	ie (O_compile_only \/ O_make_class) do
		if (Nt > 1) error(packed"multiple input -- ignoring -o", 0);
		ie (O_compile_only) do
			if (O_output_file) do
				remove(O_output_file, 1);
				rename(Names[Nt-1], O_output_file, 1);
			end
		end
		else do ! if(O_make_class)
			make_class();
		end
	end
	else do
		loadall();
		if (O_optimize) optimize(packed"txopt", [0]);
		if (O_dpfilter)
			optimize(packed"txdpf", Tmp_hack-> Tmp_hack: [0]);
		ie (O_target)
			tcode_to_native(O_target);
		else
			rename(O_tmpout, O_output_file, 1);
	end
	if (\O_keep_cdir) remove(packed"CLASSDIR", 0);
	sys.fini();
end
