!	TXPC -- Tcode Precompiler
!	Copyright (C) 2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.

#classpath "../classes";
#packstrings;

module txpc(t3x, tcode, string, xmem, iostream);

object	t[t3x], str[string], xm[xmem];

const	BUFLEN	=	1024;

const	NLABELS =	4096,
	NMARKS	=	4096;

struct	MARK = M_L, M_T, M_D;

object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Text, Data;
var	Tags[NLABELS];
var	Marks[NMARKS*MARK];
var	Dptr, Tptr;
var	Mt;
var	Opcodes;
var	Datasize, Textsize;


mwrites(s) t.write(T3X.SYSERR, s, str.length(s));


mnl() do var b::3;
	mwrites(t.newline(b));
end


fini() do
	xm.fini();
end


fail(m, n) do
	mwrites("TXPC: ");
	mwrites(m);
	if (n) do
		mwrites(": ");
		mwrites(n);
	end
	mnl();
	if (Data \= -2) fini();
	halt 1;
end


init() do var i;
	for (i=0, NLABELS) Tags[i] := -1;
	Mt := 0;
	Dptr := 0;
	Tptr := 0;
	Datasize := 32765;
	Datasize := Datasize*2;
	Textsize := Datasize;
	Data := -2;
	xm.init();
	Data := xm.alloc(Datasize);
	Text := xm.alloc(Textsize);
	if (Data < 0 \/ Text < 0) fail("insufficient memory", 0);
	In.create(T3X.SYSIN, Inbuf, BUFLEN, iostream.FREAD);
	Out.create(T3X.SYSOUT, Outbuf, BUFLEN, iostream.FWRITE);
end


addlabel(op, id) do var a, i, o, here;
	if (id >= NLABELS) fail("label too big", 0);
	if (Tags[id] \= -1) fail("duplicate label", 0);
	Tags[id] := op = tcode.ICLAB-> Tptr: Dptr;
	here := op = tcode.IDLAB-> Dptr: Tptr;
	for (i=0, Mt) if (Marks[i*MARK+M_L] = id) leave;
	if (i >= Mt) return;
	a := Marks[i*MARK+M_T];
	while (a) do
		o := a;
		a := xm.get(Text, a+1) << 8 | xm.get(Text, a);
		xm.put(Text, o, here);
		xm.put(Text, o+1, here >> 8);
	end
	a := Marks[i*MARK+M_D];
	while (a) do
		o := a;
		a := xm.get(Data, a+1) << 8 | xm.get(Data, a);
		xm.put(Data, o, here);
		xm.put(Data, o+1, here >> 8);
	end
end


addmark(id, dref) do var i, old;
	for (i=0, Mt) do
		if (Marks[i*MARK+M_L] = id) do
			ie (dref) do
				old := Marks[i*MARK+M_D];
				Marks[i*MARK+M_D] := Dptr;
			end
			else do
				old := Marks[i*MARK+M_T];
				Marks[i*MARK+M_T] := Tptr;
			end
			return old;
		end
	end
	if (Mt >= NMARKS) fail("too many marks", 0);
	Marks[Mt*MARK+M_L] := id;
	ie (dref) do
		Marks[Mt*MARK+M_D] := Dptr;
		Marks[Mt*MARK+M_T] := 0;
	end
	else do
		Marks[Mt*MARK+M_T] := Tptr;
		Marks[Mt*MARK+M_D] := 0;
	end
	Mt := Mt+1;
	return 0;
end


findlab(id, dref) do
	if (Tags[id] \= -1) return Tags[id];
	return(addmark(id, dref));
end


rdop() return In.rdch() | (In.rdch()<<8);


skip(n) do var i;
	for (i=0, n) In.rdch();
end


dgrow(n) do
	Dptr := Dptr + n;
	if (Dptr .> Datasize) fail("data segment overflow", 0);
end


tgrow(n) do
	Tptr := Tptr + n;
	if (Tptr .> Textsize) fail("text segment overflow", 0);
end


resolve() do
	var	i, op, arg, arg2, a2f, b;
	var	name::130;

	op := In.rdch();
	if (op = '#') do
		while (In.rdch() \= '\n')
			if (In.eof()) fail("bad #! header", 0);
		op := In.rdch();
	end
	if (op \= tcode.IINIT)
		fail("bad input program (magic match failed)", 0);
	while (1) do
		if (In.eof()) leave;
		if (op & 128) arg := rdop();
		a2f := 0;
		if (	op = tcode.IINCG \/ op = tcode.IINCL \/
			op = tcode.IINCI \/ op = tcode.IINIT
		) do
			arg2 := rdop();
			a2f := 1;
		end
		ie (op = tcode.ICLAB \/ op = tcode.IDLAB) do
			addlabel(op, arg);
			op := 0;
		end
		else ie (op = tcode.IDECL) do
			arg := arg-> arg<<1: 2;
			i := 0; while (i < arg)
				xm.put(Data, Dptr+i, 0);
			dgrow(arg);
			op := 0;
		end
		else ie (op = tcode.IDATA) do
			xm.put(Data, Dptr, arg);
			xm.put(Data, Dptr+1, arg>>8);
			dgrow(2);
			op := 0;
		end
		else ie (op = tcode.ICREF \/ op = tcode.IDREF) do
			arg := findlab(arg, 1);
			xm.put(Data, Dptr, arg);
			xm.put(Data, Dptr+1, arg>>8);
			dgrow(2);
			op := 0;
		end
		else ie (op = tcode.IPSTR) do
			b := Dptr;
			dgrow(arg);
			for (i=0, arg) xm.put(Data, b+i, In.rdch());
			while (1) do
				xm.put(Data, Dptr, 0);
				Dptr := Dptr+1;
				if (\(Dptr & 1)) leave;
			end
			op := 0;
		end
		else ie (op = tcode.ISTR) do
			b := Dptr;
			dgrow(arg*2);
			for (i=0, arg) do
				xm.put(Data, b, In.rdch());
				xm.put(Data, b+1, 0);
				b := b+2;
			end
			while (1) do
				xm.put(Data, Dptr, 0);
				Dptr := Dptr+1;
				if (\(Dptr & 1)) leave;
			end
			op := 0;
		end
		else ie (op = tcode.IINCG \/ op = tcode.ILDG \/
			op = tcode.ILDGV \/ op = tcode.ISAVG \/
			op = tcode.ILDLAB \/ op = tcode.ICALL \/
			op = tcode.IBRF \/ op = tcode.IBRT \/
			op = tcode.INBRF \/ op = tcode.INBRT \/
			op = tcode.IJUMP \/ op = tcode.IUNEXT \/
			op = tcode.IDNEXT
		) do
			xm.put(Text, Tptr, op);
			tgrow(1);
			arg := findlab(arg, 0);
		end
		else ie (op = tcode.IPUB) do
			arg := rdop();
			skip(arg+2);
			op := 0;
		end
		else ie (op = tcode.ILSYM \/ op = tcode.IGSYM \/
			op = tcode.IISYM
		) do
			arg := rdop();
			skip(arg+2);
			op := 0;
		end
		else ie (op = tcode.IEXT) do
			arg := rdop();
			for (i=0, arg) name::i := In.rdch();
			name::i := 0;
			fail("external reference", name);
		end
		else ie (op = tcode.IINIT) do
			if (arg & 256) fail("program already precompiled", 0);
			ie (arg = 1 \/ arg = 2)
				fail("obsolete Tcode version (1,2)", 0);
			else ie (arg = 3)
				;
			else
				fail("unsupported tcode version", 0);
			xm.put(Text, Tptr, op);
			tgrow(1);
			arg := arg | 256;
		end
		else do
			xm.put(Text, Tptr, op);
			tgrow(1);
		end
		if (op & 128) do
			xm.put(Text, Tptr, arg);
			xm.put(Text, Tptr+1, arg>>8);
			tgrow(2);
			if (a2f) do
				xm.put(Text, Tptr, arg2);
				xm.put(Text, Tptr+1, arg2>>8);
				tgrow(2);
			end
		end
		op := In.rdch();
	end
end


emit() do var i, lim;
	xm.put(Text, 3, Tptr);
	xm.put(Text, 4, Tptr>>8);
	i := 0; while (i .< Tptr) do
		Out.wrch(xm.get(Text, i));
		i := i+1;
	end
	lim := Dptr-1; while (xm.get(Data, lim)) lim := lim-1;
	lim := lim+1;
	i := 0; while (i .< lim) do
		Out.wrch(xm.get(Data, i));
		i := i+1;
	end
	Out.flush();
end


do
	init();
	resolve();
	emit();
	fini();
end
